import os
import re
from thonny import get_workbench
from thonny.ui_utils import scale

DESKTOP_SESSION = os.environ.get("DESKTOP_SESSION", "_")
CONFIGURATION_PATH = os.path.join(
    os.path.expanduser("~"), ".config/lxsession", DESKTOP_SESSION, "desktop.conf"
)
GLOBAL_CONFIGURATION_PATH = os.path.join("/etc/xdg/lxsession", DESKTOP_SESSION, "desktop.conf")


def pix():
    MAIN_BACKGROUND = "#ededed"

    res_dir = os.path.join(os.path.dirname(__file__), "res")
    scrollbar_button_settings = {}
    for direction, element_name in [
        ("up", "Vertical.Scrollbar.uparrow"),
        ("down", "Vertical.Scrollbar.downarrow"),
        ("left", "Horizontal.Scrollbar.leftarrow"),
        ("right", "Horizontal.Scrollbar.rightarrow"),
    ]:
        # load the image
        img_name = "scrollbar-button-" + direction
        for suffix in ["", "-insens"]:
            get_workbench().get_image(
                os.path.join(res_dir, img_name + suffix + ".png"), img_name + suffix
            )

        scrollbar_button_settings[element_name] = {
            "element create": (
                "image",
                img_name,
                ("!disabled", img_name),
                ("disabled", img_name + "-insens"),
            )
        }

    settings = {
        ".": {"configure": {"background": MAIN_BACKGROUND}},
        "Toolbutton": {
            "configure": {"borderwidth": 1},
            "map": {
                "relief": [("disabled", "flat"), ("hover", "groove"), ("!hover", "flat")],
                "background": [
                    ("disabled", MAIN_BACKGROUND),
                    ("!hover", MAIN_BACKGROUND),
                    ("hover", "#ffffff"),
                ],
            },
        },
        "Treeview.Heading": {
            "configure": {
                "background": "#f0f0f0",
                "foreground": "#808080",
                "relief": "flat",
                "borderwidth": 1,
            },
            "map": {"foreground": [("active", "black")]},
        },
        "TNotebook.Tab": {
            "map": {"background": [("!selected", "#d0d0d0"), ("selected", MAIN_BACKGROUND)]}
        },
        "ButtonNotebook.TNotebook.Tab": {
            "map": {
                "background": [("!selected", "#d0d0d0"), ("selected", MAIN_BACKGROUND)],
                "padding": [
                    ("selected", [scale(4), scale(2), scale(4), scale(3)]),
                    ("!selected", [scale(4), scale(2), scale(4), scale(3)]),
                ],
            }
        },
        "TScrollbar": {
            "configure": {
                "gripcount": 0,
                "borderwidth": 0,
                "padding": scale(1),
                "relief": "solid",
                "background": "#9e9e9e",
                "darkcolor": "#d6d6d6",
                "lightcolor": "#d6d6d6",
                "bordercolor": "#d6d6d6",
                "troughcolor": "#d6d6d6",
                "arrowsize": scale(1),
                "arrowcolor": "gray",
            },
            "map": {"background": [], "darkcolor": [], "lightcolor": []},
        },
        # Padding allows twaking thumb width
        "Vertical.TScrollbar": {
            "layout": [
                (
                    "Vertical.Scrollbar.trough",
                    {
                        "sticky": "ns",
                        "children": [
                            ("Vertical.Scrollbar.uparrow", {"side": "top", "sticky": ""}),
                            ("Vertical.Scrollbar.downarrow", {"side": "bottom", "sticky": ""}),
                            (
                                "Vertical.Scrollbar.padding",
                                {
                                    "sticky": "nswe",
                                    "children": [
                                        (
                                            "Vertical.Scrollbar.thumb",
                                            {"expand": 1, "sticky": "nswe"},
                                        )
                                    ],
                                },
                            ),
                        ],
                    },
                )
            ]
        },
        "Horizontal.TScrollbar": {
            "layout": [
                (
                    "Horizontal.Scrollbar.trough",
                    {
                        "sticky": "we",
                        "children": [
                            ("Horizontal.Scrollbar.leftarrow", {"side": "left", "sticky": ""}),
                            ("Horizontal.Scrollbar.rightarrow", {"side": "right", "sticky": ""}),
                            (
                                "Horizontal.Scrollbar.padding",
                                {
                                    "sticky": "nswe",
                                    "children": [
                                        (
                                            "Horizontal.Scrollbar.thumb",
                                            {"expand": 1, "sticky": "nswe"},
                                        )
                                    ],
                                },
                            ),
                        ],
                    },
                )
            ],
            "map": {
                # Make disabled Hor Scrollbar invisible
                "background": [("disabled", "#d6d6d6")],
                "troughcolor": [("disabled", "#d6d6d6")],
                "bordercolor": [("disabled", "#d6d6d6")],
                "darkcolor": [("disabled", "#d6d6d6")],
                "lightcolor": [("disabled", "#d6d6d6")],
            },
        },
        "TCombobox": {"configure": {"arrowsize": scale(10)}},
        "Menubar": {
            "configure": {
                "background": MAIN_BACKGROUND,
                "relief": "flat",
                "activebackground": "#ffffff",
                "activeborderwidth": 0,
            }
        },
        "Menu": {
            "configure": {
                "background": "#ffffff",
                "relief": "flat",
                "borderwidth": 1,
                "activeborderwidth": 0,
                # "activebackground" : bg, # updated below
                # "activeforeground" : fg,
            }
        },
        "Tooltip": {
            "configure": {
                "background": "#808080",
                "foreground": "#ffffff",
                "borderwidth": 0,
                "padx": 10,
                "pady": 10,
            }
        },
        "OPTIONS": {"configure": {"icons_in_menus": False, "shortcuts_in_tooltips": False}},
    }

    settings.update(scrollbar_button_settings)

    # try to refine settings according to system configuration
    """Note that fonts are set globally, 
    ie. all themes will later inherit these"""
    update_fonts()

    for path in [GLOBAL_CONFIGURATION_PATH, CONFIGURATION_PATH]:
        if os.path.exists(path):
            with open(path) as fp:
                for line in fp:
                    if "sGtk/ColorScheme" in line:
                        if "selected_bg_color" in line:
                            bgr = re.search(r"selected_bg_color:#([0-9a-fA-F]*)", line, re.M).group(
                                1
                            )  # @UndefinedVariable
                            color = "#" + bgr[0:2] + bgr[4:6] + bgr[8:10]
                            if is_good_color(color):
                                settings["Menu"]["configure"]["activebackground"] = color
                        if "selected_fg_color" in line:
                            fgr = re.search(r"selected_fg_color:#([0-9a-fA-F]*)", line, re.M).group(
                                1
                            )  # @UndefinedVariable
                            color = "#" + fgr[0:2] + fgr[4:6] + fgr[8:10]
                            if is_good_color(color):
                                settings["Menu"]["configure"]["activeforeground"] = color

    return settings


def is_good_color(s):
    return bool(re.match("^#[0-9a-fA-F]{6}$", s))


def pix_dark():
    update_fonts()
    return {}


def update_fonts():
    from tkinter import font

    options = {}
    for path in [GLOBAL_CONFIGURATION_PATH, CONFIGURATION_PATH]:
        if os.path.exists(path):
            with open(path) as fp:
                for line in fp:
                    if "sGtk/FontName" in line:
                        result = re.search(r"=([^0-9]*) ([0-9]*)", line, re.M)  # @UndefinedVariable
                        family = result.group(1)
                        options["size"] = int(result.group(2))

                        if re.search(r"\bBold\b", family):
                            options["weight"] = "bold"
                        else:
                            options["weight"] = "normal"

                        if re.search(r"\bItalic\b", family):
                            options["slant"] = "italic"
                        else:
                            options["slant"] = "roman"

                        options["family"] = family.replace(" Bold", "").replace(" Italic", "")

    if options:
        for name in ["TkDefaultFont", "TkMenuFont", "TkTextFont", "TkHeadingFont"]:
            font.nametofont(name).configure(**options)


def load_plugin():

    # set custom images
    if get_workbench().get_ui_mode() == "simple" and get_workbench().winfo_screenwidth() >= 1280:
        images = {
            "run-current-script": "media-playback-start48.png",
            "stop": "process-stop48.png",
            "new-file": "document-new48.png",
            "open-file": "document-open48.png",
            "save-file": "document-save48.png",
            "debug-current-script": "debug-run48.png",
            "step-over": "debug-step-over48.png",
            "step-into": "debug-step-into48.png",
            "step-out": "debug-step-out48.png",
            "run-to-cursor": "debug-run-cursor48.png",
            "tab-close": "window-close.png",
            "tab-close-active": "window-close-act.png",
            "resume": "resume48.png",
            "zoom": "zoom48.png",
            "quit": "quit48.png",
        }
    else:
        images = {
            "run-current-script": "media-playback-start.png",
            "stop": "process-stop.png",
            "new-file": "document-new.png",
            "open-file": "document-open.png",
            "save-file": "document-save.png",
            "debug-current-script": "debug-run.png",
            "step-over": "debug-step-over.png",
            "step-into": "debug-step-into.png",
            "step-out": "debug-step-out.png",
            "run-to-cursor": "debug-run-cursor.png",
            "tab-close": "window-close.png",
            "tab-close-active": "window-close-act.png",
            "resume": "resume.png",
            "zoom": "zoom.png",
            "quit": "quit.png",
        }

    res_dir = os.path.join(os.path.dirname(__file__), "res")
    theme_image_map = {}
    for image in images:
        theme_image_map[image] = os.path.join(res_dir, images[image])

    get_workbench().add_ui_theme("Raspberry Pi", "Enhanced Clam", pix, theme_image_map)
    get_workbench().add_ui_theme("Raspberry Pi Dark", "Clean Dark", pix_dark, theme_image_map)



A0 = None
A1 = None
A2 = None
A3 = None
A4 = None
A5 = None
A6 = None
A7 = None
A8 = None
A9 = None
ACCELEROMETER_INTERRUPT = None
ACCELEROMETER_SCL = None
ACCELEROMETER_SDA = None
BUTTON_A = None
BUTTON_B = None
D0 = None
D1 = None
D10 = None
D12 = None
D13 = None
D2 = None
D3 = None
D4 = None
D5 = None
D6 = None
D7 = None
D8 = None
D9 = None


def I2C():
    pass


IR_PROXIMITY = None
IR_RX = None
IR_TX = None
LIGHT = None
MICROPHONE_CLOCK = None
MICROPHONE_DATA = None
MISO = None
MOSI = None
NEOPIXEL = None
REMOTEIN = None
REMOTEOUT = None
RX = None
SCK = None
SCL = None
SDA = None
SLIDE_SWITCH = None
SPEAKER = None
SPEAKER_ENABLE = None


def SPI():
    pass


TEMPERATURE = None
TX = None


def UART():
    pass



import os.path
import tkinter as tk
from tkinter import ttk

from thonny.plugins.micropython import (
    MicroPythonProxy,
    MicroPythonConfigPage,
    add_micropython_backend,
)
from thonny import get_workbench
from thonny.ui_utils import create_url_label, show_dialog, askopenfilename
import threading
from tkinter.messagebox import showinfo
from thonny.misc_utils import list_volumes


class CircuitPythonProxy(MicroPythonProxy):
    @property
    def known_usb_vids_pids(self):
        """Copied from https://github.com/mu-editor/mu/blob/master/mu/modes/adafruit.py"""
        return {
            (0x239A, 0x8015),  # Adafruit Feather M0 CircuitPython
            (0x239A, 0x8023),  # Adafruit Feather M0 Express CircuitPython
            (0x239A, 0x801B),  # Adafruit Feather M0 Express CircuitPython
            (0x239A, 0x8014),  # Adafruit Metro M0 CircuitPython
            (0x239A, 0x8019),  # Adafruit CircuitPlayground Express CircuitPython
            (0x239A, 0x801D),  # Adafruit Gemma M0
            (0x239A, 0x801F),  # Adafruit Trinket M0
            (0x239A, 0x8012),  # Adafruit ItsyBitsy M0
            (0x239A, 0x8021),  # Adafruit Metro M4
            (0x239A, 0x8025),  # Adafruit Feather RadioFruit
            (0x239A, 0x8026),  # Adafruit Feather M4
            (0x239A, 0x8028),  # Adafruit pIRKey M0
            (0x239A, 0x802A),  # Adafruit Feather 52840
            (0x239A, 0x802C),  # Adafruit Itsy M4
            (0x239A, 0x802E),  # Adafruit CRICKit M0
        }


class CircuitPythonConfigPage(MicroPythonConfigPage):
    def _get_usb_driver_url(self):
        return "https://learn.adafruit.com/welcome-to-circuitpython/installing-circuitpython"


def load_plugin():
    add_micropython_backend(
        "CircuitPython", CircuitPythonProxy, _("CircuitPython (generic)"), CircuitPythonConfigPage
    )



DIRTYPE = "dir"


class FileSection:
    ""

    def read():
        pass

    def readinto():
        pass

    def skip():
        pass


REGTYPE = "file"
TAR_HEADER = None


class TarFile:
    ""

    def extractfile():
        pass

    def next():
        pass


class TarInfo:
    ""


def roundup():
    pass


uctypes = None



BADDY = None
BA_DING = None
BIRTHDAY = None
BLUES = None
CHASE = None
DADADADUM = None
ENTERTAINER = None
FUNERAL = None
FUNK = None
JUMP_DOWN = None
JUMP_UP = None
NYAN = None
ODE = None
POWER_DOWN = None
POWER_UP = None
PRELUDE = None
PUNCHLINE = None
PYTHON = None
RINGTONE = None
WAWAWAWAA = None
WEDDING = None


def get_tempo():
    pass


def pitch():
    pass


def play():
    pass


def reset():
    pass


def set_tempo():
    pass


def stop():
    pass



# -*- coding: utf-8 -*-
from __future__ import print_function

"""
This file is run by CPythonProxy

(Why separate file for launching? I want to have clean global scope 
in toplevel __main__ module (because that's where user scripts run), but backend's global scope 
is far from clean. 
I could also do python -c "from backend import VM: VM().mainloop()", but looks like this 
gives relative __file__-s on imported modules.) 
"""

if __name__ == "__main__":
    import sys
    import platform

    if platform.system() == "Darwin":
        import os

        try:
            os.getcwd()
        except Exception:
            print(
                "\nNB! Potential problems detected, see\nhttps://github.com/thonny/thonny/wiki/MacOSX#catalina\n",
                file=sys.stderr,
            )

    if not sys.version_info > (3, 5):
        print(
            "Thonny only supports Python 3.5 and later.\n"
            + "Choose another interpreter from Tools => Options => Interpreter",
            file=sys.stderr,
        )
        exit()

    import logging
    import os.path

    # remove script dir from path
    sys.path.pop(0)

    # import thonny relative to this script (not from current interpreter path)
    import importlib.util

    spec = importlib.util.spec_from_file_location(
        "thonny", os.path.join(os.path.dirname(__file__), "__init__.py")
    )
    thonny = importlib.util.module_from_spec(spec)
    sys.modules["thonny"] = thonny
    assert spec.loader is not None
    spec.loader.exec_module(thonny)

    THONNY_USER_DIR = os.environ["THONNY_USER_DIR"]
    # set up logging
    logger = logging.getLogger("thonny")
    logger.propagate = False
    logFormatter = logging.Formatter("%(levelname)s: %(message)s")
    file_handler = logging.FileHandler(
        os.path.join(THONNY_USER_DIR, "backend.log"), encoding="UTF-8", mode="w"
    )
    file_handler.setFormatter(logFormatter)
    file_handler.setLevel(logging.INFO)
    logger.addHandler(file_handler)

    # Don't litter user stderr with thonny logging
    # TODO: Can I somehow send the log to front-end's stderr?
    """
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setLevel(logging.INFO);
    stream_handler.setFormatter(logFormatter)
    logger.addHandler(stream_handler)
    """

    logger.setLevel(logging.INFO)

    import faulthandler

    fault_out = open(os.path.join(THONNY_USER_DIR, "backend_faults.log"), mode="w")
    faulthandler.enable(fault_out)

    # Disable blurry scaling in Windows
    thonny.set_dpi_aware()

    from thonny.backend import VM  # @UnresolvedImport

    VM().mainloop()



from thonny.plugins.micropython import (
    MicroPythonProxy,
    MicroPythonConfigPage,
    add_micropython_backend,
    list_serial_ports_with_descriptions,
)
from thonny import get_workbench, get_runner, ui_utils
import os
import subprocess
from thonny.ui_utils import (
    SubprocessDialog,
    CommonDialog,
    CommonDialogEx,
    ems_to_pixels,
    show_dialog,
)
from thonny.running import get_frontend_python, get_interpreter_for_subprocess
from time import sleep

import tkinter as tk
from tkinter import ttk, filedialog, messagebox
from collections import OrderedDict
import tkinter
from thonny.plugins.micropython.serial_connection import SerialConnection
from thonny.misc_utils import running_on_mac_os, construct_cmd_line
import time


class ESPProxy(MicroPythonProxy):
    pass


class ESP8266Proxy(ESPProxy):
    description = "MicroPython on ESP8266"
    config_page_constructor = "ESP8266"

    @property
    def known_usb_vids_pids(self):
        return super().known_usb_vids_pids | {
            # eg. Adafruit Feather Huzzah
            (0x10C4, 0xEA60),  # Silicon Labs CP210x USB to UART Bridge,
            (0x1A86, 0x7523),  # USB-SERIAL CH340,
        }

    def _get_api_stubs_path(self):
        return os.path.join(os.path.dirname(__file__), "esp8266_api_stubs")


class ESP32Proxy(ESPProxy):
    @property
    def known_usb_vids_pids(self):
        return super().known_usb_vids_pids | {
            # eg. ESP-WROOM-32
            (0x10C4, 0xEA60)  # Silicon Labs CP210x USB to UART Bridge
        }

    def _get_api_stubs_path(self):
        return os.path.join(os.path.dirname(__file__), "esp32_api_stubs")


class ESPConfigPage(MicroPythonConfigPage):
    def __init__(self, master, chip, firmware_start_address):
        self._chip = chip
        self._firmware_start_address = firmware_start_address
        super().__init__(master)

    def _get_flashing_frame(self):
        frame = super()._get_flashing_frame()
        return frame

    def _has_flashing_dialog(self):
        return True

    def _open_flashing_dialog(self):
        dlg = ESPFlashingDialog(get_workbench(), self._chip, self._firmware_start_address)
        ui_utils.show_dialog(dlg)

    @property
    def allow_webrepl(self):
        return False


class ESP8266ConfigPage(ESPConfigPage):
    def __init__(self, master):
        super().__init__(master, "esp8266", "0x0")


class ESP32ConfigPage(ESPConfigPage):
    def __init__(self, master):
        super().__init__(master, "esp32", "0x1000")


class ESPFlashingDialog(CommonDialogEx):
    def __init__(self, master, chip, start_address, initial_port_desc=""):
        super().__init__(master)

        self.title("Install %s firmware with esptool" % chip.upper())

        self._chip = chip
        self._start_address = start_address
        self._esptool_command = self._get_esptool_command()
        if not self._esptool_command:
            messagebox.showerror(
                "Can't find esptool",
                "esptool not found.\n"
                + "Install it via 'Tools => Manage plug-ins'\n"
                + "or using your OP-system package manager.",
            )
            return self._close()

        self.main_frame.columnconfigure(2, weight=1)

        epadx = ems_to_pixels(2)
        ipadx = ems_to_pixels(0.5)
        epady = epadx
        ipady = ipadx

        # Port
        port_label = ttk.Label(self.main_frame, text="Port")
        port_label.grid(row=1, column=1, sticky="w", padx=(epadx, 0), pady=(epady, 0))

        self._port_desc_variable = tk.StringVar(value=initial_port_desc)
        self._port_combo = ttk.Combobox(
            self.main_frame, exportselection=False, textvariable=self._port_desc_variable, values=[]
        )
        self._port_combo.state(["!disabled", "readonly"])
        self._port_combo.grid(row=1, column=2, sticky="nsew", padx=ipadx, pady=(epady, 0))

        port_reload_button = ttk.Button(self.main_frame, text="Reload", command=self._reload_ports)
        port_reload_button.grid(row=1, column=3, sticky="ew", padx=(0, epadx), pady=(epady, 0))

        # Firmware
        firmware_label = ttk.Label(self.main_frame, text="Firmware")
        firmware_label.grid(row=2, column=1, sticky="w", padx=(epadx, 0), pady=(ipady, 0))

        self._firmware_entry = ttk.Entry(self.main_frame, width=65)
        self._firmware_entry.grid(row=2, column=2, sticky="nsew", padx=ipadx, pady=(ipady, 0))

        browse_button = ttk.Button(self.main_frame, text="Browse...", command=self._browse)
        browse_button.grid(row=2, column=3, sticky="we", padx=(0, epadx), pady=(ipady, 0))

        # Erase
        self._erase_variable = tk.BooleanVar(value=True)
        self._erase_checkbutton = ttk.Checkbutton(
            self.main_frame, text="Erase flash before installing", variable=self._erase_variable
        )
        self._erase_checkbutton.grid(
            row=3, column=1, columnspan=3, sticky="w", padx=(epadx, 0), pady=(ipady, 0)
        )

        # Buttons
        install_button = ttk.Button(self.main_frame, text="Install", command=self._install)
        install_button.grid(row=4, column=1, columnspan=2, sticky="e", padx=ipadx, pady=(0, epady))

        cancel_button = ttk.Button(self.main_frame, text="Close", command=self._close)
        cancel_button.grid(
            row=4, column=3, columnspan=1, sticky="we", padx=(0, epadx), pady=(0, epady)
        )

        self._reload_ports()

    def _get_esptool_command(self):
        try:
            import esptool

            return [get_interpreter_for_subprocess(), "-u", "-m", "esptool"]
        except ImportError:
            import shutil

            result = shutil.which("esptool")
            if result:
                return [result]
            else:
                result = shutil.which("esptool.py")
                if result:
                    return [result]
                else:
                    return None

    def _reload_ports(self):
        pairs = list_serial_ports_with_descriptions()
        self._ports_by_desc = OrderedDict(pairs)
        self._port_combo.configure(values=list(self._ports_by_desc.keys()))

    def _browse(self):
        path = ui_utils.askopenfilename(filetypes=[("bin-files", ".bin"), ("all files", ".*")])
        if path:
            self._firmware_entry.delete(0, "end")
            self._firmware_entry.insert(0, path)

    def _check_connection(self, port):
        proxy = get_runner().get_backend_proxy()
        if isinstance(proxy, MicroPythonProxy):
            # Most likely it is using the same port
            proxy.disconnect()
            time.sleep(1.5)

        # Maybe another program is connected
        # or the user doesn't have sufficient permissions?
        try:
            conn = SerialConnection(port, 115200, skip_reader=True)
            conn.close()

            return True
        except Exception as e:
            messagebox.showerror(
                "Can't connect", str(e), master=None if running_on_mac_os() else self
            )
            return False

    def _install(self):
        if not self._port_desc_variable.get():
            messagebox.showerror("Select port", "Please select port")
            return

        port = self._ports_by_desc[self._port_desc_variable.get()]

        firmware_path = self._firmware_entry.get()
        if not os.path.exists(firmware_path):
            messagebox.showerror("Bad firmware path", "Can't find firmware, please check path")
            return

        erase_command = self._esptool_command + [
            "--chip",
            self._chip,
            "--port",
            port,
            "erase_flash",
        ]

        write_command = self._esptool_command + [
            "--chip",
            self._chip,
            "--port",
            port,
            "write_flash",
            self._start_address,
            firmware_path,
        ]

        if not self._check_connection(port):
            return

        # unknown problem with first dialog appearing at 0,0
        # self.update_idletasks()
        if self.winfo_screenwidth() >= 1024:
            min_left = ems_to_pixels(15)
            min_top = ems_to_pixels(5)
        else:
            min_left = 0
            min_top = 0

        def long_desc(cmd):
            return "Command:\n%s\n\nOutput:\n" % construct_cmd_line(cmd)

        self.update_idletasks()
        if self._erase_variable.get():
            proc = self._create_subprocess(erase_command)
            dlg = SubprocessDialog(
                self,
                proc,
                "Erasing flash",
                long_description=long_desc(erase_command),
                autoclose=True,
            )
            show_dialog(dlg, master=self, min_left=min_left, min_top=min_top)
            if dlg.cancelled or dlg.returncode:
                return

        proc = self._create_subprocess(write_command)
        dlg = SubprocessDialog(
            self,
            proc,
            "Installing firmware",
            long_description=long_desc(write_command),
            autoclose=False,
        )
        show_dialog(dlg, master=self, min_left=min_left, min_top=min_top)

    def _create_subprocess(self, cmd):
        return subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True
        )

    def _close(self):
        self.closed = True
        self.destroy()


def load_plugin():
    add_micropython_backend("ESP8266", ESP8266Proxy, "MicroPython (ESP8266)", ESP8266ConfigPage)
    add_micropython_backend("ESP32", ESP32Proxy, "MicroPython (ESP32)", ESP32ConfigPage)

    def upload_micropython():
        proxy = get_runner().get_backend_proxy()
        proxy.select_and_upload_micropython()

    def upload_micropython_enabled():
        proxy = get_runner().get_backend_proxy()
        return getattr(proxy, "micropython_upload_enabled", False) and isinstance(proxy, ESPProxy)

    def erase_flash():
        proxy = get_runner().get_backend_proxy()
        proxy.erase_flash()

    def erase_flash_enabled():
        return (
            isinstance(get_runner().get_backend_proxy(), ESPProxy)
            and get_runner().get_backend_proxy().micropython_upload_enabled
        )

    get_workbench().add_command(
        "uploadmicropythonesp",
        "device",
        "Install MicroPython to ESP8266/ESP32 ...",
        upload_micropython,
        upload_micropython_enabled,
        group=40,
    )

    get_workbench().add_command(
        "erasespflash",
        "device",
        "Erase ESP8266/ESP32 flash",
        erase_flash,
        tester=erase_flash_enabled,
        group=40,
    )



# -*- coding: utf-8 -*-

from _tkinter import TclError
import logging
import os.path
import re
from tkinter import ttk
import traceback

from thonny import get_runner, get_workbench, memory, roughparse, ui_utils, running
from thonny.codeview import PythonText, get_syntax_options_for_tag
from thonny.common import InlineCommand, ToplevelCommand, ToplevelResponse
from thonny.misc_utils import construct_cmd_line, parse_cmd_line, running_on_mac_os, shorten_repr
from thonny.tktextext import index2line, TextFrame, TweakableText
from thonny.ui_utils import (
    EnhancedTextWithLogging,
    scrollbar_style,
    select_sequence,
    TextMenu,
    create_tooltip,
    show_dialog,
    lookup_style_option,
    CommonDialog,
    tr_btn,
)
import tkinter as tk
import webbrowser
from thonny.running import EDITOR_CONTENT_TOKEN

_CLEAR_SHELL_DEFAULT_SEQ = select_sequence("<Control-l>", "<Command-k>")

# NB! Don't add parens without refactoring split procedure!
OUTPUT_SPLIT_REGEX = re.compile(r"(\x1B\[[0-?]*[ -/]*[@-~]|[\a\b\r])")
NUMBER_SPLIT_REGEX = re.compile(r"((?<!\w)[-+]?[0-9]*\.?[0-9]+\b)")
SIMPLE_URL_SPLIT_REGEX = re.compile(
    r"(https?:\/\/[\w\/.:\-\?#=%]+[\w\/]|data:image\/[a-z]+;base64,[A-Za-z0-9\/=\+]+)"
)

INT_REGEX = re.compile(r"\d+")
ANSI_COLOR_NAMES = {
    "0": "black",
    "1": "red",
    "2": "green",
    "3": "yellow",
    "4": "blue",
    "5": "magenta",
    "6": "cyan",
    "7": "white",
    "9": "default",
}


class ShellView(tk.PanedWindow):
    def __init__(self, master):
        super().__init__(
            master,
            orient="horizontal",
            sashwidth=lookup_style_option("Sash", "sashthickness", 4),
            background=lookup_style_option("TPanedWindow", "background"),
            borderwidth=0,
        )

        main_frame = tk.Frame(self)
        self.add(main_frame, minsize=100)

        self.vert_scrollbar = ttk.Scrollbar(
            main_frame, orient=tk.VERTICAL, style=scrollbar_style("Vertical")
        )
        self.vert_scrollbar.grid(row=1, column=2, sticky=tk.NSEW)
        get_workbench().add_command(
            "clear_shell",
            "edit",
            _("Clear shell"),
            self.clear_shell,
            default_sequence=_CLEAR_SHELL_DEFAULT_SEQ,
            group=200,
        )

        get_workbench().set_default("shell.max_lines", 1000)
        get_workbench().set_default("shell.squeeze_threshold", 1000)
        get_workbench().set_default("shell.tty_mode", True)

        self.text = ShellText(
            main_frame,
            self,
            font="EditorFont",
            # foreground="white",
            # background="#666666",
            highlightthickness=0,
            # highlightcolor="LightBlue",
            borderwidth=0,
            yscrollcommand=self.set_scrollbar,
            padx=0,
            pady=0,
            insertwidth=2,
            height=10,
            undo=True,
        )

        get_workbench().event_generate("ShellTextCreated", text_widget=self.text)
        get_workbench().bind("TextInsert", self.text_inserted, True)
        get_workbench().bind("TextDelete", self.text_deleted, True)

        self.text.grid(row=1, column=1, sticky=tk.NSEW)
        self.vert_scrollbar["command"] = self.text.yview
        main_frame.columnconfigure(1, weight=1)
        main_frame.rowconfigure(1, weight=1)

        self.notice = ttk.Label(self, text="", background="#ffff99", padding=3)

        self.init_plotter()
        self.menu = ShellMenu(self.text, self)

    def init_plotter(self):
        self.plotter = None
        get_workbench().set_default("view.show_plotter", False)
        get_workbench().set_default("view.shell_sash_position", 400)

        self.plotter_visibility_var = get_workbench().get_variable("view.show_plotter")

        def can_toggle():
            return self.winfo_ismapped()

        get_workbench().add_command(
            "toggle_plotter",
            "view",
            _("Plotter"),
            self.toggle_plotter,
            can_toggle,
            flag_name="view.show_plotter",
            group=11,
        )

        self.update_plotter_visibility(True)

    def toggle_plotter(self):
        self.plotter_visibility_var.set(not self.plotter_visibility_var.get())
        self.update_plotter_visibility()

    def update_plotter_visibility(self, initializing_shell_view=False):
        if self.plotter_visibility_var.get():
            self.show_plotter(initializing_shell_view)
        else:
            self.hide_plotter()

    def show_plotter(self, initializing_shell_view=False):
        if not initializing_shell_view:
            get_workbench().show_view("ShellView", True)

        if self.plotter is None:
            self.plotter = PlotterCanvas(self, self.text)

        if not self.plotter.winfo_ismapped():
            self.add(self.plotter, minsize=100)

        self.sash_place(0, get_workbench().get_option("view.shell_sash_position"), 0)

        running.io_animation_required = True
        self.update_plotter()

    def hide_plotter(self):
        if self.plotter is None or not self.plotter.winfo_ismapped():
            return
        else:
            self.remove(self.plotter)
            running.io_animation_required = False

    def set_notice(self, text):
        if text is None:
            self.notice.grid_forget()
        else:
            self.notice["text"] = text
            if not self.notice.winfo_ismapped():
                self.notice.grid(row=0, column=1, columnspan=2, sticky="nsew", pady=(0, 1))
                # height of the text was reduced so adjust the scrolling
                # self.update()
                self.text.see("end")

    def print_error(self, txt):
        self.text._insert_text_directly(txt, ("io", "stderr"))

    def focus_set(self):
        self.text.focus_set()

    def submit_python_code(self, cmd_line):
        self.text.submit_command(cmd_line, ())

    def submit_magic_command(self, cmd_line):
        if isinstance(cmd_line, list):
            cmd_line = construct_cmd_line(cmd_line)

        if not cmd_line.endswith("\n"):
            cmd_line += "\n"

        self.text.submit_command(cmd_line, ("magic",))

    def restart(self):
        self.text.restart()

    def clear_shell(self):
        self.text._clear_shell()

    def has_pending_input(self):
        return self.text.has_pending_input()

    def report_exception(self, prelude=None, conclusion=None):
        if prelude is not None:
            self.text.direct_insert("end", prelude + "\n", ("stderr",))

        self.text.direct_insert("end", traceback.format_exc() + "\n", ("stderr",))

        if conclusion is not None:
            self.text.direct_insert("end", conclusion + "\n", ("stderr",))

    def set_scrollbar(self, *args):
        self.vert_scrollbar.set(*args)
        self.update_plotter()

    def text_deleted(self, event):
        if event.text_widget == self.text:
            self.update_plotter()

    def text_inserted(self, event):
        if (
            event.text_widget == self.text
            and "\n" in event.text
            # only when scrollbar doesn't move, because otherwise
            # the update gets triggered by scrollbar anyway
            and self.vert_scrollbar.get() == (0.0, 1.0)
        ):
            self.update_plotter()

    def update_plotter(self):
        if self.plotter is not None and self.plotter.winfo_ismapped():
            self.plotter.update_plot()

    def update_tabs(self):
        self.text.update_tabs()

    def resize_plotter(self):
        if len(self.panes()) > 1 and self.text.winfo_width() > 5:
            get_workbench().set_option("view.shell_sash_position", self.sash_coord(0)[0])


class ShellMenu(TextMenu):
    def __init__(self, target, view):
        self.view = view
        TextMenu.__init__(self, target)

    def add_extra_items(self):
        self.add_separator()
        self.add_command(label=_("Clear"), command=self.text._clear_shell)

        def toggle_from_menu():
            # I don't like that Tk menu toggles checbutton variable
            # automatically before calling the handler.
            # So I revert the toggle before calling the actual handler.
            # This way the handler doesn't have to worry whether it
            # needs to toggle the variable or not, and it can choose to
            # decline the toggle.
            self.view.plotter_visibility_var.set(not self.view.plotter_visibility_var.get())
            self.view.toggle_plotter()

        self.add_checkbutton(
            label=_("Show Plotter"),
            command=toggle_from_menu,
            variable=self.view.plotter_visibility_var,
        )

    def selection_is_read_only(self):
        return not self.text.selection_is_writable()


class BaseShellText(EnhancedTextWithLogging, PythonText):
    """Passive version of ShellText. Used also for preview"""

    def __init__(self, master, view=None, cnf={}, **kw):
        self.view = view
        kw["tabstyle"] = "wordprocessor"
        super().__init__(master, cnf, **kw)

        self._command_history = (
            []
        )  # actually not really history, because each command occurs only once
        self._command_history_current_index = None

        # logs of IO events for current toplevel block
        # (enables undoing and redoing the events)
        self._applied_io_events = []
        self._queued_io_events = []
        self._images = set()

        self._ansi_foreground = None
        self._ansi_background = None
        self._ansi_inverse = False
        self._ansi_intensity = None
        self._ansi_italic = False
        self._ansi_underline = False
        self._ansi_conceal = False
        self._ansi_strikethrough = False
        self._io_cursor_offset = 0
        self._squeeze_buttons = set()

        self.update_tty_mode()

        self.bind("<Up>", self._arrow_up, True)
        self.bind("<Down>", self._arrow_down, True)
        self.bind("<KeyPress>", self._text_key_press, True)
        self.bind("<KeyRelease>", self._text_key_release, True)

        prompt_font = tk.font.nametofont("BoldEditorFont")
        x_padding = 4
        io_vert_spacing = 10
        io_indent = 16 + x_padding
        self.io_indent = io_indent
        code_indent = prompt_font.measure(">>> ") + x_padding

        self.tag_configure("command", lmargin1=code_indent, lmargin2=code_indent)
        self.tag_configure(
            "io", lmargin1=io_indent, lmargin2=io_indent, rmargin=io_indent, font="IOFont"
        )
        self.update_margin_color()

        self.tag_configure("after_io_or_value", spacing1=io_vert_spacing)
        self.tag_configure("before_io", spacing3=io_vert_spacing)

        self.tag_configure("prompt", lmargin1=x_padding, lmargin2=x_padding)
        self.tag_configure("value", lmargin1=x_padding, lmargin2=x_padding)
        self.tag_configure("restart_line", wrap="none", lmargin1=x_padding, lmargin2=x_padding)

        self.tag_configure("welcome", lmargin1=x_padding, lmargin2=x_padding)

        # Underline on the font looks better than underline on the tag,
        # therefore Shell doesn't use configured "hyperlink" style directly
        hyperlink_opts = get_syntax_options_for_tag("hyperlink").copy()
        if hyperlink_opts.get("underline"):
            hyperlink_opts["font"] = "UnderlineIOFont"
            del hyperlink_opts["underline"]
        self.tag_configure("io_hyperlink", **hyperlink_opts)

        # create 3 marks: input_start shows the place where user entered but not-yet-submitted
        # input starts, output_end shows the end of last output,
        # output_insert shows where next incoming program output should be inserted
        self.mark_set("input_start", "end-1c")
        self.mark_gravity("input_start", tk.LEFT)

        self.mark_set("output_end", "end-1c")
        self.mark_gravity("output_end", tk.LEFT)

        self.mark_set("output_insert", "end-1c")
        self.mark_gravity("output_insert", tk.RIGHT)

        self.active_object_tags = set()

        self.update_tabs()

        self.tag_raise("io_hyperlink")
        self.tag_raise("underline")
        self.tag_raise("strikethrough")
        self.tag_raise("intense_io")
        self.tag_raise("italic_io")
        self.tag_raise("intense_italic_io")
        self.tag_raise("sel")

    def submit_command(self, cmd_line, tags):
        # assert get_runner().is_waiting_toplevel_command()
        self.delete("input_start", "end")
        self.insert("input_start", cmd_line, tags)
        self.see("end")
        self.mark_set("insert", "end")
        self._try_submit_input()

    def _handle_input_request(self, msg):
        self._ensure_visible()
        self.focus_set()
        self.mark_set("insert", "end")
        self.tag_remove("sel", "1.0", tk.END)
        self._try_submit_input()  # try to use leftovers from previous request
        self.see("end")

    def _handle_program_output(self, msg):
        # Discard but not too often, as toplevel response will discard anyway
        if int(float(self.index("end"))) > get_workbench().get_option("shell.max_lines") + 100:
            self._discard_old_content()

        self._ensure_visible()
        self._append_to_io_queue(msg.data, msg.stream_name)

        if not self._applied_io_events:
            # this is first line of io, add padding below command line
            self.tag_add("before_io", "output_insert -1 line linestart")

        self._update_visible_io(None)

    def _handle_toplevel_response(self, msg: ToplevelResponse) -> None:
        if msg.get("error"):
            self._insert_text_directly(msg["error"] + "\n", ("toplevel", "stderr"))
            self._ensure_visible()

        if "user_exception" in msg:
            self._show_user_exception(msg["user_exception"])
            self._ensure_visible()

        welcome_text = msg.get("welcome_text")
        if welcome_text and welcome_text:
            preceding = self.get("output_insert -1 c", "output_insert")
            if preceding.strip() and not preceding.endswith("\n"):
                self._insert_text_directly("\n")
            self._insert_text_directly(welcome_text, ("welcome",))

        if "value_info" in msg:
            num_stripped_question_marks = getattr(msg, "num_stripped_question_marks", 0)
            if num_stripped_question_marks > 0:
                # show the value in object inspector
                get_workbench().event_generate("ObjectSelect", object_id=msg["value_info"].id)
            else:
                # show the value in shell
                value_repr = shorten_repr(msg["value_info"].repr, 10000)
                if value_repr != "None":
                    if get_workbench().in_heap_mode():
                        value_repr = memory.format_object_id(msg["value_info"].id)
                    object_tag = "object_" + str(msg["value_info"].id)
                    self._insert_text_directly(value_repr + "\n", ("toplevel", "value", object_tag))
                    if running_on_mac_os():
                        sequence = "<Command-Button-1>"
                    else:
                        sequence = "<Control-Button-1>"
                    self.tag_bind(
                        object_tag,
                        sequence,
                        lambda _: get_workbench().event_generate(
                            "ObjectSelect", object_id=msg["value_info"].id
                        ),
                    )

                    self.active_object_tags.add(object_tag)

        self.mark_set("output_end", self.index("end-1c"))
        self._discard_old_content()
        self._update_visible_io(None)
        self._reset_ansi_attributes()
        self._io_cursor_offset = 0
        self._insert_prompt()
        self._try_submit_input()  # Trying to submit leftover code (eg. second magic command)
        self.see("end")

        # import os
        # import psutil
        # process = psutil.Process(os.getpid())
        # print("MEM", process.memory_info().rss // (1024*1024))

    def _handle_fancy_debugger_progress(self, msg):
        if msg.in_present or msg.io_symbol_count is None:
            self._update_visible_io(None)
        else:
            self._update_visible_io(msg.io_symbol_count)

    def _get_squeeze_threshold(self):
        return get_workbench().get_option("shell.squeeze_threshold")

    def _append_to_io_queue(self, data, stream_name):
        if self.tty_mode:
            # Make sure ANSI CSI codes are stored as separate events
            # TODO: try to complete previously submitted incomplete code
            parts = re.split(OUTPUT_SPLIT_REGEX, data)
            for part in parts:
                if part:  # split may produce empty string in the beginning or start
                    # split the data so that very long lines separated
                    for block in re.split("(.{%d,})" % (self._get_squeeze_threshold() + 1), part):
                        if block:
                            self._queued_io_events.append((block, stream_name))
        else:
            self._queued_io_events.append((data, stream_name))

    def _update_visible_io(self, target_num_visible_chars):
        current_num_visible_chars = sum(map(lambda x: len(x[0]), self._applied_io_events))

        if (
            target_num_visible_chars is not None
            and target_num_visible_chars < current_num_visible_chars
        ):
            # hard to undo complex renderings (squeezed texts and ANSI codes)
            # easier to clean everything and start again
            self._queued_io_events = self._applied_io_events + self._queued_io_events
            self._applied_io_events = []
            self.direct_delete("command_io_start", "output_end")
            current_num_visible_chars = 0
            self._reset_ansi_attributes()

        while self._queued_io_events and current_num_visible_chars != target_num_visible_chars:
            data, stream_name = self._queued_io_events.pop(0)

            if target_num_visible_chars is not None:
                leftover_count = current_num_visible_chars + len(data) - target_num_visible_chars

                if leftover_count > 0:
                    # add suffix to the queue
                    self._queued_io_events.insert(0, (data[-leftover_count:], stream_name))
                    data = data[:-leftover_count]

            self._apply_io_event(data, stream_name)
            current_num_visible_chars += len(data)

        self.mark_set("output_end", self.index("end-1c"))
        self.see("end")

    def _apply_io_event(self, data, stream_name, extra_tags=set()):
        if not data:
            return

        original_data = data

        if self.tty_mode and re.match(OUTPUT_SPLIT_REGEX, data):
            if data == "\a":
                get_workbench().bell()
            elif data == "\b":
                self._change_io_cursor_offset(-1)
            elif data == "\r":
                self._change_io_cursor_offset("line")
            elif data.endswith("D") or data.endswith("C"):
                self._change_io_cursor_offset_csi(data)
            elif stream_name == "stdout":
                # According to https://github.com/tartley/colorama/blob/master/demos/demo04.py
                # codes sent to stderr shouldn't affect later output in stdout
                # It makes sense, but Ubuntu terminal does not confirm it.
                # For now I'm just trimming stderr color codes
                self._update_ansi_attributes(data)

        else:
            tags = extra_tags | {"io", stream_name}
            if stream_name == "stdout" and self.tty_mode:
                tags |= self._get_ansi_tags()

            non_url_length = len(data)
            for url_match in SIMPLE_URL_SPLIT_REGEX.finditer(data):
                non_url_length -= url_match.end() - url_match.start()

            if non_url_length > self._get_squeeze_threshold() and "\n" not in data:
                self._io_cursor_offset = 0  # ignore the effect of preceding \r and \b
                button_text = data[:40] + " …"
                btn = tk.Label(
                    self,
                    text=button_text,
                    # width=len(button_text),
                    cursor="arrow",
                    borderwidth=2,
                    relief="raised",
                    font="IOFont",
                )
                btn.bind("<1>", lambda e: self._show_squeezed_text(btn), True)
                btn.contained_text = data
                btn.tags = tags
                self._squeeze_buttons.add(btn)
                create_tooltip(btn, "%d characters squeezed. " % len(data) + "Click for details.")

                # TODO: refactor
                # (currently copied from insert_text_directly)
                self.mark_gravity("input_start", tk.RIGHT)
                self.mark_gravity("output_insert", tk.RIGHT)

                self.window_create("output_insert", window=btn)
                for tag_name in tags:
                    self.tag_add(tag_name, "output_insert -1 chars")
                data = ""

            elif self._io_cursor_offset < 0:
                overwrite_len = min(len(data), -self._io_cursor_offset)

                if 0 <= data.find("\n") < overwrite_len:
                    overwrite_len = data.find("\n")

                overwrite_data = data[:overwrite_len]
                self.direct_insert(
                    "output_insert -%d chars" % -self._io_cursor_offset, overwrite_data, tuple(tags)
                )
                del_start = self.index("output_insert -%d chars" % -self._io_cursor_offset)
                del_end = self.index(
                    "output_insert -%d chars" % (-self._io_cursor_offset - overwrite_len)
                )
                self.direct_delete(del_start, del_end)

                # compute leftover data to be printed normally
                data = data[overwrite_len:]

                if "\n" in data:
                    # cursor offset doesn't apply on new line
                    self._io_cursor_offset = 0
                else:
                    # offset becomes closer to 0
                    self._io_cursor_offset + overwrite_len

            elif self._io_cursor_offset > 0:
                # insert spaces before actual data
                # NB! Print without formatting tags
                self._insert_text_directly(" " * self._io_cursor_offset, ("io", stream_name))
                self._io_cursor_offset = 0

            if data:
                # if any data is still left, then this should be output normally
                self._insert_text_directly(data, tuple(tags))

        self._applied_io_events.append((original_data, stream_name))

    def _show_squeezed_text(self, button):
        dlg = SqueezedTextDialog(self, button)
        show_dialog(dlg)

    def _change_io_cursor_offset_csi(self, marker):
        ints = re.findall(INT_REGEX, marker)
        if len(ints) != 1:
            logging.warn("bad CSI cursor positioning: %s", marker)
            # do nothing
            return

        try:
            delta = int(ints[0])
        except ValueError:
            logging.warn("bad CSI cursor positioning: %s", marker)
            return

        if marker.endswith("D"):
            delta = -delta

        self._change_io_cursor_offset(delta)

    def _change_io_cursor_offset(self, delta):
        line = self.get("output_insert linestart", "output_insert")
        if delta == "line":
            self._io_cursor_offset = -len(line)
        else:
            self._io_cursor_offset += delta
            if self._io_cursor_offset < -len(line):
                # cap
                self._io_cursor_offset = -len(line)

    def _reset_ansi_attributes(self):
        self._ansi_foreground = None
        self._ansi_background = None
        self._ansi_inverse = False
        self._ansi_intensity = None
        self._ansi_italic = False
        self._ansi_underline = False
        self._ansi_conceal = False
        self._ansi_strikethrough = False

    def _update_ansi_attributes(self, marker):
        if not marker.endswith("m"):
            # ignore
            return

        codes = re.findall(INT_REGEX, marker)
        if not codes:
            self._reset_ansi_attributes()

        while codes:
            code = codes.pop(0)

            if code == "0":
                self._reset_ansi_attributes()
            elif code in ["1", "2"]:
                self._ansi_intensity = code
            elif code == "3":
                self._ansi_italic = True
            elif code == "4":
                self._ansi_underline = True
            elif code == "7":
                self._ansi_inverse = True
            elif code == "8":
                self._ansi_conceal = True
            elif code == "9":
                self._ansi_strikethrough = True
            elif code == "22":
                self._ansi_intensity = None
            elif code == "23":
                self._ansi_italic = False
            elif code == "24":
                self._ansi_underline = False
            elif code == "27":
                self._ansi_inverse = False
            elif code == "28":
                self._ansi_conceal = False
            elif code == "29":
                self._ansi_strikethrough = False
            if code in [
                "30",
                "31",
                "32",
                "33",
                "34",
                "35",
                "36",
                "37",
                "90",
                "91",
                "92",
                "93",
                "94",
                "95",
                "96",
                "97",
            ]:
                self._ansi_foreground = code
            elif code == "39":
                self._ansi_foreground = None
            elif code in [
                "40",
                "41",
                "42",
                "43",
                "44",
                "45",
                "46",
                "47",
                "100",
                "101",
                "102",
                "103",
                "104",
                "105",
                "106",
                "107",
            ]:
                self._ansi_background = code
            elif code == "49":
                self._ansi_background = None
            elif code in ["38", "48"]:
                # multipart code, ignore for now,
                # but make sure all arguments are ignored
                if not codes:
                    # nothing follows, ie. invalid code
                    break
                mode = codes.pop(0)
                if mode == "5":
                    # 256-color code, just ignore for now
                    if not codes:
                        break
                    codes = codes[1:]
                elif mode == "2":
                    # 24-bit code, ignore
                    if len(codes) < 3:
                        # invalid code
                        break
                    codes = codes[3:]
            else:
                # ignore other codes
                pass

    def _get_ansi_tags(self):
        result = set()

        if self._ansi_foreground:
            fg = ANSI_COLOR_NAMES[self._ansi_foreground[-1]]
            if self._ansi_intensity == "1" or self._ansi_foreground[0] == "9":
                fg = "bright_" + fg
            elif self._ansi_intensity == "2":
                fg = "dim_" + fg
        else:
            fg = "fore"
            if self._ansi_intensity == "1":
                fg = "bright_" + fg
            elif self._ansi_intensity == "2":
                fg = "dim_" + fg

        if self._ansi_background:
            bg = ANSI_COLOR_NAMES[self._ansi_background[-1]]
            if self._ansi_background.startswith("10"):
                bg = "bright_" + bg
        else:
            bg = "back"

        if self._ansi_inverse:
            result.add(fg + "_bg")
            result.add(bg + "_fg")
        else:
            if fg != "fore":
                result.add(fg + "_fg")
            if bg != "back":
                result.add(bg + "_bg")

        if self._ansi_intensity == "1" and self._ansi_italic:
            result.add("intense_italic_io")
        elif self._ansi_intensity == "1":
            result.add("intense_io")
        elif self._ansi_italic:
            result.add("italic_io")

        if self._ansi_underline:
            result.add("underline")

        if self._ansi_strikethrough:
            result.add("strikethrough")

        return result

    def _insert_prompt(self):
        # if previous output didn't put a newline, then do it now
        if not self.index("output_insert").endswith(".0"):
            self._insert_text_directly("\n", ("io",))

        prompt_tags = ("toplevel", "prompt")

        # if previous line has value or io then add little space
        prev_line = self.index("output_insert - 1 lines")
        prev_line_tags = self.tag_names(prev_line)
        if "io" in prev_line_tags or "value" in prev_line_tags:
            prompt_tags += ("after_io_or_value",)

        self._insert_text_directly(">>> ", prompt_tags)
        self.edit_reset()

    def _ensure_visible(self):
        if self.winfo_ismapped():
            return

        focused_view = get_workbench().focus_get()
        get_workbench().show_view("ShellView")
        if focused_view is not None:
            focused_view.focus()

    def update_tabs(self):
        tab_chars = 8
        tab_pixels = tk.font.nametofont("IOFont").measure("n" * tab_chars)

        offset = self.io_indent
        tabs = [offset]
        for _ in range(20):
            offset += tab_pixels
            tabs.append(offset)

        self.tag_configure("io", tabs=tabs, tabstyle="wordprocessor")

    def restart(self):
        self._insert_text_directly(
            # "\n============================== RESTART ==============================\n",
            "\n" + "─" * 200 + "\n",
            # "\n" + "═"*200 + "\n",
            ("magic", "restart_line"),
        )

    def intercept_insert(self, index, txt, tags=()):
        # pylint: disable=arguments-differ
        if self._editing_allowed() and self._in_current_input_range(index):
            # self._print_marks("before insert")
            # I want all marks to stay in place
            self.mark_gravity("input_start", tk.LEFT)
            self.mark_gravity("output_insert", tk.LEFT)

            if get_runner().is_waiting_toplevel_command():
                tags = tags + ("toplevel", "command")
            else:
                tags = tags + ("io", "stdin")

            EnhancedTextWithLogging.intercept_insert(self, index, txt, tags)

            if not get_runner().is_waiting_toplevel_command():
                if not self._applied_io_events:
                    # tag preceding command line differently
                    self.tag_add("before_io", "input_start -1 lines linestart")

                self._try_submit_input()

            self.see("insert")
        else:
            get_workbench().bell()

    def intercept_delete(self, index1, index2=None, **kw):
        if index1 == "sel.first" and index2 == "sel.last" and not self.has_selection():
            return

        if (
            self._editing_allowed()
            and self._in_current_input_range(index1)
            and (index2 is None or self._in_current_input_range(index2))
        ):
            self.direct_delete(index1, index2, **kw)
        else:
            get_workbench().bell()

    def selection_is_writable(self):
        try:
            if not self.has_selection():
                return self._in_current_input_range(self.index("insert"))
            else:
                return self._in_current_input_range(
                    self.index("sel.first")
                ) and self._in_current_input_range(self.index("sel.last"))
        except TclError:
            return True

    def perform_return(self, event):
        if get_runner().is_running():
            # if we are fixing the middle of the input string and pressing ENTER
            # then we expect the whole line to be submitted not linebreak to be inserted
            # (at least that's how IDLE works)
            self.mark_set("insert", "end")  # move cursor to the end

            # Do the return without auto indent
            EnhancedTextWithLogging.perform_return(self, event)

            self._try_submit_input()

        elif get_runner().is_waiting_toplevel_command():
            # Same with editin middle of command, but only if it's a single line command
            whole_input = self.get("input_start", "end-1c")  # asking the whole input
            if "\n" not in whole_input and self._code_is_ready_for_submission(whole_input):
                self.mark_set("insert", "end")  # move cursor to the end
                # Do the return without auto indent
                EnhancedTextWithLogging.perform_return(self, event)
            else:
                # Don't want auto indent when code is ready for submission
                source = self.get("input_start", "insert")
                tail = self.get("insert", "end")

                if self._code_is_ready_for_submission(source + "\n", tail):
                    # No auto-indent
                    EnhancedTextWithLogging.perform_return(self, event)
                else:
                    # Allow auto-indent
                    PythonText.perform_return(self, event)

            self._try_submit_input()

        return "break"

    def on_secondary_click(self, event=None):
        super().on_secondary_click(event)
        if self.view:
            self.view.menu.tk_popup(event.x_root, event.y_root)

    def _in_current_input_range(self, index):
        try:
            return self.compare(index, ">=", "input_start")
        except Exception:
            return False

    def _insert_text_directly(self, txt, tags=()):
        def _insert(txt, tags):
            if txt != "":
                self.direct_insert("output_insert", txt, tags)

        def _insert_and_highlight_urls(txt, tags):
            parts = SIMPLE_URL_SPLIT_REGEX.split(txt)
            for i, part in enumerate(parts):
                if i % 2 == 0:
                    _insert(part, tags)
                else:
                    if part.startswith("data:image/"):
                        token = ";base64,"
                        data = part[part.index(token) + len(token) :]
                        try:
                            img = tk.PhotoImage(data=data)
                            self._images.add(img)  # to avoit it being gc-d"""
                            self.image_create("output_insert", image=img)
                            for tag in tags:
                                self.tag_add(tag, "output_insert -1 chars")
                        except TclError:
                            _insert(part, tags + ("io_hyperlink",))
                    else:
                        _insert(part, tags + ("io_hyperlink",))

        # I want the insertion to go before marks
        # self._print_marks("before output")
        self.mark_gravity("input_start", tk.RIGHT)
        self.mark_gravity("output_insert", tk.RIGHT)
        tags = tuple(tags)

        # Make stacktrace clickable
        if "stderr" in tags or "error" in tags:
            # show lines pointing to source lines as hyperlinks
            for line in txt.splitlines(True):
                parts = re.split(r"(File .* line \d+.*)$", line, maxsplit=1)
                if len(parts) == 3 and "<pyshell" not in line:
                    _insert(parts[0], tags)
                    _insert(parts[1], tags + ("io_hyperlink",))
                    _insert(parts[2], tags)
                else:
                    parts = re.split(r"(\'[^\']+\.pyw?\')", line, flags=re.IGNORECASE)
                    if len(parts) == 3 and os.path.exists(os.path.expanduser(parts[1][1:-1])):
                        match = re.search(r"\S", line)
                        _insert(line[: match.start()], tags)
                        _insert(line[match.start() :], tags + ("io_hyperlink",))
                    else:
                        _insert_and_highlight_urls(line, tags)
        else:
            _insert_and_highlight_urls(txt, tags)

        # self._print_marks("after output")
        # output_insert mark will move automatically because of its gravity

    def has_pending_input(self):
        pending = self.get("input_start", "end-1c")
        return bool(pending)

    def _try_submit_input(self):
        # see if there is already enough inputted text to submit
        input_text = self.get("input_start", "insert")
        tail = self.get("insert", "end")

        # user may have pasted more text than necessary for this request
        submittable_text = self._extract_submittable_input(input_text, tail)

        if submittable_text is not None:
            if get_runner().is_waiting_toplevel_command():
                # clean up the tail
                if len(tail) > 0:
                    assert tail.strip() == ""
                    self.delete("insert", "end-1c")

            # leftover text will be kept in widget, waiting for next request.
            start_index = self.index("input_start")
            end_index = self.index("input_start+{0}c".format(len(submittable_text)))

            # apply correct tags (if it's leftover then it doesn't have them yet)
            if get_runner().is_running():
                self.tag_add("io", start_index, end_index)
                self.tag_add("stdin", start_index, end_index)
            else:
                self.tag_add("toplevel", start_index, end_index)
                self.tag_add("command", start_index, end_index)

            # update start mark for next input range
            self.mark_set("input_start", end_index)

            # Move output_insert mark after the requested_text
            # Leftover input, if any, will stay after output_insert,
            # so that any output that will come in before
            # next input request will go before leftover text
            self.mark_set("output_insert", end_index)

            # remove tags from leftover text
            for tag in ("io", "stdin", "toplevel", "command"):
                # don't remove magic, because otherwise I can't know it's auto
                self.tag_remove(tag, end_index, "end")

            self._submit_input(submittable_text)

    def _editing_allowed(self):
        return get_runner() is not None

    def _extract_submittable_input(self, input_text, tail):

        if get_runner().is_waiting_toplevel_command():
            if input_text.endswith("\n"):
                if input_text.strip().startswith("%") or input_text.strip().startswith("!"):
                    # if several magic command are submitted, then take only first
                    return input_text[: input_text.index("\n") + 1]
                elif self._code_is_ready_for_submission(input_text, tail):
                    return input_text
                else:
                    return None
            else:
                return None

        elif get_runner().is_running():
            i = 0
            while True:
                if i >= len(input_text):
                    return None
                elif input_text[i] == "\n":
                    return input_text[: i + 1]
                else:
                    i += 1

        return None

    def _code_is_ready_for_submission(self, source, tail=""):
        # Ready to submit if ends with empty line
        # or is complete single-line code

        if tail.strip() != "":
            return False

        # First check if it has unclosed parens, unclosed string or ending with : or \
        parser = roughparse.RoughParser(self.indent_width, self.tabwidth)
        parser.set_str(source.rstrip() + "\n")
        if parser.get_continuation_type() != roughparse.C_NONE or parser.is_block_opener():
            return False

        # Multiline compound statements need to end with empty line to be considered
        # complete.
        lines = source.splitlines()
        # strip starting empty and comment lines
        while len(lines) > 0 and (lines[0].strip().startswith("#") or lines[0].strip() == ""):
            lines.pop(0)

        compound_keywords = ["if", "while", "for", "with", "try", "def", "class", "async", "await"]
        if len(lines) > 0:
            first_word = lines[0].strip().split()[0]
            if first_word in compound_keywords and not source.replace(" ", "").replace(
                "\t", ""
            ).endswith("\n\n"):
                # last line is not empty
                return False

        return True

    def _submit_input(self, text_to_be_submitted):
        logging.debug(
            "SHELL: submitting %r in state %s", text_to_be_submitted, get_runner().get_state()
        )
        if get_runner().is_waiting_toplevel_command():
            # register in history and count
            if text_to_be_submitted in self._command_history:
                self._command_history.remove(text_to_be_submitted)
            self._command_history.append(text_to_be_submitted)

            # meaning command selection is not in process
            self._command_history_current_index = None

            self.update_tty_mode()

            cmd_line = text_to_be_submitted.strip()
            try:
                if cmd_line.startswith("%"):
                    parts = cmd_line.split(" ", maxsplit=1)
                    if len(parts) == 2:
                        args_str = parts[1].strip()
                    else:
                        args_str = ""
                    argv = parse_cmd_line(cmd_line[1:])
                    command_name = argv[0]
                    cmd_args = argv[1:]

                    if len(cmd_args) >= 2 and cmd_args[0] == "-c":
                        # move source argument to source attribute
                        source = cmd_args[1]
                        cmd_args = [cmd_args[0]] + cmd_args[2:]
                        if source == EDITOR_CONTENT_TOKEN:
                            source = (
                                get_workbench().get_editor_notebook().get_current_editor_content()
                            )
                    else:
                        source = None

                    get_workbench().event_generate("MagicCommand", cmd_line=text_to_be_submitted)
                    get_runner().send_command(
                        ToplevelCommand(
                            command_name,
                            args=cmd_args,
                            args_str=args_str,
                            cmd_line=cmd_line,
                            tty_mode=self.tty_mode,
                            source=source,
                        )
                    )
                elif cmd_line.startswith("!"):
                    argv = parse_cmd_line(cmd_line[1:])
                    get_workbench().event_generate("SystemCommand", cmd_line=text_to_be_submitted)
                    get_runner().send_command(
                        ToplevelCommand(
                            "execute_system_command",
                            argv=argv,
                            cmd_line=cmd_line,
                            tty_mode=self.tty_mode,
                        )
                    )
                else:
                    get_runner().send_command(
                        ToplevelCommand(
                            "execute_source", source=text_to_be_submitted, tty_mode=self.tty_mode
                        )
                    )

                # remember the place where the output of this command started
                self.mark_set("command_io_start", "output_insert")
                self.mark_gravity("command_io_start", "left")
                # discard old io events
                self._applied_io_events = []
                self._queued_io_events = []
            except Exception:
                get_workbench().report_exception()
                self._insert_prompt()

            get_workbench().event_generate("ShellCommand", command_text=text_to_be_submitted)
        else:
            assert get_runner().is_running()
            get_runner().send_program_input(text_to_be_submitted)
            get_workbench().event_generate("ShellInput", input_text=text_to_be_submitted)
            self._applied_io_events.append((text_to_be_submitted, "stdin"))

    def _arrow_up(self, event):
        if not get_runner().is_waiting_toplevel_command():
            return None

        if not self._in_current_input_range("insert"):
            return None

        insert_line = index2line(self.index("insert"))
        input_start_line = index2line(self.index("input_start"))
        if insert_line != input_start_line:
            # we're in the middle of a multiline command
            return None

        if len(self._command_history) == 0 or self._command_history_current_index == 0:
            # can't take previous command
            return "break"

        if self._command_history_current_index is None:
            self._command_history_current_index = len(self._command_history) - 1
        else:
            self._command_history_current_index -= 1

        cmd = self._command_history[self._command_history_current_index]
        if cmd[-1] == "\n":
            cmd = cmd[:-1]  # remove the submission linebreak
        self._propose_command(cmd)
        return "break"

    def _arrow_down(self, event):
        if not get_runner().is_waiting_toplevel_command():
            return None

        if not self._in_current_input_range("insert"):
            return None

        insert_line = index2line(self.index("insert"))
        last_line = index2line(self.index("end-1c"))
        if insert_line != last_line:
            # we're in the middle of a multiline command
            return None

        if (
            len(self._command_history) == 0
            or self._command_history_current_index is None
            or self._command_history_current_index >= len(self._command_history) - 1
        ):
            # can't take next command
            self._command_history_current_index = len(self._command_history)
            self._propose_command("")
            return "break"

        if self._command_history_current_index is None:
            self._command_history_current_index = len(self._command_history) - 1
        else:
            self._command_history_current_index += 1

        self._propose_command(
            self._command_history[self._command_history_current_index].strip("\n")
        )
        return "break"

    def _propose_command(self, cmd_line):
        self.delete("input_start", "end")
        self.intercept_insert("input_start", cmd_line)
        self.see("insert")

    def _text_key_press(self, event):
        # Ctrl should underline values
        # TODO: this underline may confuse, when user is just copying on pasting
        # try to add this underline only when mouse is over the value

        # TODO: take theme into account
        """
        if event.keysym in ("Control_L", "Control_R", "Command"):  # TODO: check in Mac
            self.tag_configure("value", foreground="DarkBlue", underline=1)
        """

    def _text_key_release(self, event):
        # Remove value underlining
        # TODO: take theme into account
        """
        if event.keysym in ("Control_L", "Control_R", "Command"):  # TODO: check in Mac
            self.tag_configure("value", foreground="DarkBlue", underline=0)
        """

    def _clear_shell(self):
        end_index = self.index("output_end")
        self._clear_content(end_index)

    def compute_smart_home_destination_index(self):
        """Is used by EnhancedText"""

        if self._in_current_input_range("insert"):
            # on input line, go to just after prompt
            return "input_start"
        else:
            return super().compute_smart_home_destination_index()

    def _hyperlink_enter(self, event):
        self.config(cursor="hand2")

    def _hyperlink_leave(self, event):
        self.config(cursor="")

    def _handle_hyperlink(self, event):
        try:
            line = self.get("insert linestart", "insert lineend")
            # Python stacktrace
            matches = list(re.finditer(r'File "(?P<file>[^"]+)", line (?P<line>\d+)', line))
            if not matches:
                # Friendly traceback
                matches = list(
                    re.finditer(
                        r"\b(?P<line>\d+)\b.+'(?P<file>[^\']+\.pyw?)'", line, flags=re.IGNORECASE
                    )
                )

            if len(matches) == 1:
                filename = os.path.expanduser(matches[0].group("file"))
                lineno = int(matches[0].group("line"))
                if os.path.exists(filename) and os.path.isfile(filename):
                    # TODO: better use events instead direct referencing
                    get_workbench().get_editor_notebook().show_file(
                        filename, lineno, set_focus=False
                    )
            else:
                r = self.tag_prevrange("io_hyperlink", "@%d,%d" % (event.x, event.y))
                if r and len(r) == 2:
                    url = self.get(r[0], r[1])
                    if SIMPLE_URL_SPLIT_REGEX.match(url):
                        webbrowser.open(url)

        except Exception:
            traceback.print_exc()

    def _show_user_exception(self, user_exception):
        fr_tr = user_exception.get("friendly_traceback", None)
        if fr_tr is not None:
            self._format_friendly_traceback(fr_tr)
            return

        for line, frame_id, *_ in user_exception["items"]:

            tags = ("io", "stderr")
            if frame_id is not None:
                frame_tag = "frame_%d" % frame_id

                def handle_frame_click(event, frame_id=frame_id):
                    get_runner().send_command(InlineCommand("get_frame_info", frame_id=frame_id))
                    return "break"

                # TODO: put first line with frame tag and rest without
                tags += (frame_tag,)
                self.tag_bind(frame_tag, "<ButtonRelease-1>", handle_frame_click, True)

            self._insert_text_directly(line, tags)

    def _format_friendly_traceback(self, s):
        tags = ("io", "stderr")

        lines = s.splitlines(True)
        for line in lines:
            self._insert_text_directly(line, tags)

    def _discard_old_content(self):
        max_lines = max(get_workbench().get_option("shell.max_lines"), 0)
        proposed_cut = self.index("end -%d lines linestart" % max_lines)
        if proposed_cut == "1.0":
            return

        # would this keep current block intact?
        next_prompt = self.tag_nextrange("prompt", proposed_cut, "end")
        if not next_prompt:
            pass  # TODO: disable stepping back

        self._clear_content(proposed_cut)

    def _clear_content(self, cut_idx):
        proposed_cut_float = float(cut_idx)
        for btn in list(self._squeeze_buttons):
            btn_pos = float(self.index(btn))
            if btn_pos < proposed_cut_float:
                self._squeeze_buttons.remove(btn)
                # looks like the widgets are not fully GC-d.
                # At least avoid leaking big chunks of texts
                btn.contained_text = None
                btn.destroy()

        self.direct_delete("0.1", cut_idx)

    def _invalidate_current_data(self):
        """
        Grayes out input & output displayed so far
        """
        end_index = self.index("output_end")

        self.tag_add("inactive", "1.0", end_index)
        self.tag_remove("value", "1.0", end_index)

        while len(self.active_object_tags) > 0:
            self.tag_remove(self.active_object_tags.pop(), "1.0", "end")

    def get_lines_above_viewport_bottom(self, tag_name, n):
        end_index = self.index("@%d,%d lineend" % (self.winfo_height(), self.winfo_height()))
        start_index = self.index(end_index + " -50 lines")

        result = ""
        while True:
            r = self.tag_nextrange(tag_name, start_index, end_index)
            if not r:
                break
            result += self.get(r[0], r[1])
            start_index = r[1]

        return result

    def update_tty_mode(self):
        self.tty_mode = get_workbench().get_option("shell.tty_mode")

    def set_syntax_options(self, syntax_options):
        super().set_syntax_options(syntax_options)
        self.update_margin_color()

    def update_margin_color(self):
        if ui_utils.get_tk_version_info() >= (8, 6, 6):
            self.tag_configure("io", lmargincolor=get_syntax_options_for_tag("TEXT")["background"])


class ShellText(BaseShellText):
    def __init__(self, master, view, cnf={}, **kw):
        super().__init__(master, view, cnf=cnf, **kw)
        self.bindtags(self.bindtags() + ("ShellText",))

        self.tag_bind("io_hyperlink", "<ButtonRelease-1>", self._handle_hyperlink)
        self.tag_bind("io_hyperlink", "<Enter>", self._hyperlink_enter)
        self.tag_bind("io_hyperlink", "<Leave>", self._hyperlink_leave)

        get_workbench().bind("InputRequest", self._handle_input_request, True)
        get_workbench().bind("ProgramOutput", self._handle_program_output, True)
        get_workbench().bind("ToplevelResponse", self._handle_toplevel_response, True)
        get_workbench().bind("DebuggerResponse", self._handle_fancy_debugger_progress, True)


class SqueezedTextDialog(CommonDialog):
    def __init__(self, master, button):
        super().__init__(master)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.button = button
        self.content = button.contained_text
        self.shell_text = master

        padding = 20

        mainframe = ttk.Frame(self)
        mainframe.grid(row=0, column=0, sticky="nsew")
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(2, weight=1)

        explanation_label = ttk.Label(
            mainframe,
            text=_(
                "For performance reasons, Shell avoids showing "
                + "very long lines in full (see Tools => Options => Shell).\n"
                + "Here you can interact with the original text fragment."
            ),
        )
        explanation_label.grid(row=0, column=0, sticky="nsew", padx=padding, pady=padding)

        self._wrap_var = tk.BooleanVar(False)
        self.wrap_checkbox = ttk.Checkbutton(
            mainframe,
            text=_("Wrap text (may be slow)"),
            variable=self._wrap_var,
            onvalue=True,
            offvalue=False,
            command=self._on_wrap_changed,
        )
        self.wrap_checkbox.grid(row=1, padx=padding, pady=(0, padding // 2), sticky="w")

        self.text_frame = TextFrame(
            mainframe,
            text_class=TweakableText,
            height=10,
            width=80,
            relief="sunken",
            borderwidth=1,
            wrap="none",
        )
        self.text_frame.grid(row=2, column=0, padx=padding, sticky="nsew")
        self.text_frame.text.insert("1.0", button.contained_text)
        self.text_frame.text.set_read_only(True)

        button_frame = ttk.Frame(mainframe)
        button_frame.grid(row=3, column=0, padx=padding, pady=padding, sticky="nswe")
        button_frame.columnconfigure(2, weight=1)

        copy_caption = tr_btn("Copy to clipboard")
        copy_button = ttk.Button(
            button_frame, text=copy_caption, width=len(copy_caption), command=self._on_copy
        )
        copy_button.grid(row=0, column=1, sticky="w", padx=(0, padding))

        expand_caption = tr_btn("Expand in Shell")
        expand_button = ttk.Button(
            button_frame, text=expand_caption, width=len(expand_caption), command=self._on_expand
        )
        expand_button.grid(row=0, column=2, sticky="e", padx=padding)

        close_button = ttk.Button(button_frame, text=tr_btn("Close"), command=self._on_close)
        close_button.grid(row=0, column=3, sticky="e")

        self.bind("<Escape>", self._on_close, True)
        self.protocol("WM_DELETE_WINDOW", self._on_close)
        self.title(_("Squeezed text (%d characters)") % len(self.content))

    def _on_wrap_changed(self):
        if self._wrap_var.get():
            self.text_frame.text.configure(wrap="word")
        else:
            self.text_frame.text.configure(wrap="none")

    def _on_expand(self):
        index = self.shell_text.index(self.button)
        self.shell_text.direct_delete(index, index + " +1 chars")
        self.shell_text.direct_insert(index, self.content, tuple(self.button.tags))
        self.destroy()

        # looks like the widgets are not fully GC-d.
        # At least avoid leaking big chunks of texts
        self.button.contained_text = None
        self.button.destroy()

    def _on_copy(self):
        self.clipboard_clear()
        self.clipboard_append(self.content)

    def _on_close(self, event=None):
        self.destroy()


class PlotterCanvas(tk.Canvas):
    def __init__(self, master, text):
        self.master = master
        self.background = get_syntax_options_for_tag("TEXT")["background"]
        self.foreground = get_syntax_options_for_tag("TEXT")["foreground"]
        super().__init__(
            master,
            background=self.background,
            borderwidth=0,
            height=10000,  # size of the virtual drawing area
            width=10000,
            highlightthickness=0,
        )
        self.text = text

        self.x_scale = None
        self.x_scale = None
        self.range_start = -1
        self.range_end = 2
        self.range_block_size = 0
        self.value_range = 2
        self.last_legend = None
        self.font = tk.font.nametofont("TkDefaultFont")
        self.linespace = self.font.metrics("linespace")
        self.y_padding = self.linespace
        self.x_padding_left = -1  # makes sharper cut for partly hidden line
        self.x_padding_right = self.linespace
        self.fresh_range = True

        self.colors = [
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#8c564b",
            "#e377c2",
            "#7f7f7f",
            "#bcbd22",
            "#17becf",
        ]
        self.range_block_sizes = [
            0.1,
            0.25,
            0.5,
            1,
            5,
            10,
            25,
            50,
            100,
            250,
            500,
            1000,
            2500,
            5000,
            10000,
        ]
        self.bind("<Configure>", self.on_resize, True)
        self.bind("<Button-1>", self.reset_range, True)

        self.create_close_button()

        get_workbench().bind("SyntaxThemeChanged", self.reload_theme_options, True)

    def create_close_button(self):
        self.close_img = get_workbench().get_image("tab-close")
        self.close_active_img = get_workbench().get_image("tab-close-active")

        self.close_rect = self.create_rectangle(
            self.winfo_width() - self.close_img.width() - self.linespace,
            self.linespace / 2,
            self.winfo_width(),
            self.linespace / 2 + self.close_img.height(),
            fill=self.background,
            width=0,
            tags=("close",),
        )

        self.close_button = self.create_image(
            self.winfo_width() - self.linespace / 2,
            self.linespace / 2,
            anchor="ne",
            image=self.close_img,
            activeimage=self.close_active_img,
            tags=("close",),
        )

        self.tag_bind("close", "<1>", self.on_close)

    def update_close_button(self):
        self.coords(
            self.close_rect,
            self.winfo_width() - self.close_img.width() - self.linespace / 1.5,
            self.linespace / 2,
            self.winfo_width() - self.linespace / 2,
            self.linespace / 2 + self.close_img.height(),
        )
        self.coords(self.close_button, self.winfo_width() - self.linespace / 2, self.linespace / 2)

    def on_close(self, event):
        self.master.toggle_plotter()

    def reset_range(self, event=None):
        self.fresh_range = True

    def get_num_steps(self):
        return 30

    def update_plot(self, force_clean=False):
        data_lines = []
        bottom_index = self.text.index(
            "@%d,%d" % (self.text.winfo_width(), self.text.winfo_height())
        )
        bottom_lineno = int(float(bottom_index))

        for i in range(bottom_lineno - self.get_num_steps(), bottom_lineno + 1):
            line_start_index = "%d.0" % i
            if i < 1 or "stdout" not in self.text.tag_names(line_start_index):
                data_lines.append(([], []))
            else:
                content = self.text.get(line_start_index, line_start_index + " lineend")
                data_lines.append(self.extract_pattern_and_numbers(content))

        # data_lines need to be transposed
        segments_by_color = []
        for i in range(100):
            segments = list(self.extract_series_segments(data_lines, i))
            if segments:
                segments_by_color.append(segments)
            else:
                break

        self.delete("segment")

        self.update_range(segments_by_color, force_clean)
        segment_count = self.draw_segments(segments_by_color)
        self.update_legend(data_lines, force_clean)

        self.delete("info")
        if segment_count == 0:
            info_text = (
                _("Plotter visualizes series of\n" + "numbers printed to the Shell.")
                + "\n\n"
                + _("See Help for details.")
            )

            self.create_text_with_background(
                self.winfo_width() / 2,
                self.winfo_height() / 2,
                text=info_text,
                anchor="center",
                justify="center",
                tags=("info",),
            )
            # self.delete("guide", "tick", "legend")
            # self.range_start = 0
            # self.range_end = 0
            self.tag_raise("info")

        self.fresh_range = False

    def update_legend(self, data_lines, force_clean=False):
        legend = None
        i = len(data_lines) - 2  # one before last
        while i >= 0:
            legend = data_lines[i][0]
            if legend and legend == data_lines[i + 1][0]:
                # found last legend, which covers at least 2 consecutive points
                break
            i -= 1

        if self.last_legend == legend and not force_clean:
            # just make sure it remains topmost
            self.tag_raise("legend")
            return

        self.delete("legend")

        if legend is None:
            return

        # add horizontal padding
        # legend[0] = " " + legend[0]
        # legend[-1] = legend[-1] + " "

        marker = "●"  # "●" "•"
        marker_width = self.font.measure(marker)
        full_text_width = self.font.measure(marker.join(legend))

        y = self.winfo_height() - self.linespace // 2
        x = self.winfo_width() - full_text_width - self.linespace

        self.create_rectangle(
            x - self.linespace // 4,
            y - self.linespace,
            x + full_text_width + self.linespace // 4,
            y,
            fill=self.background,
            width=0,
            tags=("legend",),
        )

        for i, part in enumerate(legend):
            if i > 0:
                self.create_text(
                    x,
                    y,
                    text=marker,
                    anchor="sw",
                    fill=self.colors[i - 1 % len(self.colors)],
                    tags=("legend",),
                )
                x += marker_width

            self.create_text(x, y, text=part, anchor="sw", tags=("legend",), fill=self.foreground)
            x += self.font.measure(part)

        self.last_legend = legend

    def draw_segments(self, segments_by_color):
        count = 0
        for color, segments in enumerate(segments_by_color):
            for pos, nums in segments:
                self.draw_segment(color, pos, nums)
                count += 1

        # raise certain elements above segments
        self.tag_raise("tick")
        self.tag_raise("close")
        return count

    def draw_segment(self, color, pos, nums):

        x = self.x_padding_left + pos * self.x_scale

        args = []
        for num in nums:
            y = self.y_padding + (self.range_end - num) * self.y_scale
            args.extend([x, y])
            x += self.x_scale

        self.create_line(
            *args,
            width=2,
            fill=self.colors[color % len(self.colors)],
            tags=("segment",),
            # arrow may be confusing
            # and doesn't play nice with distinguising between
            # scrollback view and fresh_range view
            # arrow="last",
            # arrowshape=(3,5,3)
        )
        # self.current_segment_ids.append(line_id)

    def update_range(self, segments_by_color, clean):
        if not segments_by_color:
            return

        range_start = 9999999999
        range_end = -9999999999

        # if new block is using 3/4 of the width,
        # then don't consider old block's values anymore
        interest_position = 0
        for start_pos, nums in reversed(segments_by_color[0]):
            if start_pos < self.get_num_steps() / 10:
                interest_position = start_pos
                break

        assert isinstance(interest_position, int)
        for segments in segments_by_color:
            for start_pos, nums in segments:
                if start_pos >= interest_position:
                    range_start = min(range_start, *nums)
                    range_end = max(range_end, *nums)

        if interest_position == 0 and not self.fresh_range:
            # meaning we still care about old line's values
            range_start = min(range_start, self.range_start)
            range_end = max(range_end, self.range_end)

        if range_end == range_start:
            range_end += 1

        if (
            not clean
            and not self.fresh_range
            and self.x_scale is not None
            and range_end == self.range_end
            and range_start == self.range_start
        ):
            # don't recompute as nothing was changed
            return

        value_range = range_end - range_start
        range_block_size = value_range // 4
        # prefer round blocks
        for size in self.range_block_sizes:
            if size * 4 >= value_range:
                range_block_size = size
                break

        # extend to range block boundary
        if range_end % range_block_size != 0:
            range_end -= range_end % -range_block_size

        if range_start % range_block_size != 0:
            range_start -= range_start % range_block_size

        # not sure about these assertions when using floats
        # assert range_start % range_block_size == 0
        # assert range_end % range_block_size == 0, "range_end: %s, bs: %s" % (range_end, range_block_size)

        # remember
        self.range_start = range_start
        self.range_end = range_end
        self.value_range = range_end - range_start
        self.range_block_size = range_block_size

        available_height = self.winfo_height() - 2 * self.y_padding
        available_width = self.winfo_width() - self.x_padding_left - self.x_padding_right
        num_steps = self.get_num_steps()

        self.x_scale = available_width / (num_steps - 1)
        self.y_scale = available_height / self.value_range

        self.update_guides_and_ticks()

    def update_guides_and_ticks(self):
        self.delete("guide", "tick")
        value = self.range_start
        while value <= self.range_end:
            y = self.y_padding + (self.range_end - value) * self.y_scale

            # guide
            self.create_line(
                0, y, self.winfo_width(), y, tags=("guide",), dash=(2, 2), fill="#aaaaaa"
            )

            # tick
            if value == int(value):
                value = int(value)

            caption = " " + str(value) + " "
            self.create_text_with_background(
                self.linespace // 2, y, caption, anchor="w", tags=("tick",)
            )
            value += self.range_block_size

    def extract_pattern_and_numbers(self, line):
        parts = NUMBER_SPLIT_REGEX.split(line)
        if len(parts) < 2:
            return ([], [])

        assert len(parts) % 2 == 1

        pattern = []
        numbers = []
        for i in range(0, len(parts), 2):
            pattern.append(parts[i])

        for i in range(1, len(parts), 2):
            numbers.append(float(parts[i]))

        return (pattern, numbers)

    def extract_series_segments(self, data_lines, series_nr):
        """Yields numbers which form connected multilines on graph
        Each segment is pair of starting position and numbers"""
        segment = (0, [])
        prev_pattern = None
        for i, (pattern, nums) in enumerate(data_lines):
            if len(nums) <= series_nr or pattern != prev_pattern:
                # break the segment
                if len(segment[1]) > 1:
                    yield segment
                segment = (i, [])

            if len(nums) > series_nr:
                segment[1].append(nums[series_nr])

            prev_pattern = pattern

        if len(segment[1]) > 1:
            yield segment

    def create_text_with_background(
        self, x, y, text, anchor="w", justify="left", background=None, tags=()
    ):
        if background is None:
            background = self.background

        width = 0
        lines = text.splitlines()
        for line in lines:
            width = max(width, self.font.measure(line))

        height = len(lines) * self.linespace

        rect_x = x
        rect_y = y
        if anchor == "center":
            rect_x = x - width / 2
            rect_y = y - height / 2
        elif anchor == "w":
            rect_y = y - height / 2
        else:
            "TODO:"

        self.create_rectangle(
            rect_x, rect_y, rect_x + width, rect_y + height, fill=background, width=0, tags=tags
        )
        self.create_text(
            x, y, anchor=anchor, text=text, tags=tags, fill=self.foreground, justify=justify
        )

    def reload_theme_options(self, event):
        self.background = get_syntax_options_for_tag("TEXT")["background"]
        self.foreground = get_syntax_options_for_tag("TEXT")["foreground"]
        self.configure(background=self.background)
        self.itemconfig(self.close_rect, fill=self.background)
        self.update_plot(True)

    def on_resize(self, event):
        if self.winfo_width() > 10:
            get_workbench().set_option("view.plotter_width", self.winfo_width())
        self.update_plot(True)
        self.update_close_button()
        self.master.resize_plotter()



# Stubs for turtle (generated from Python 3.7)

import tkinter as TK
from typing import Any, Optional, Union, overload, Tuple, Dict, List

class Vec2D(tuple):
    def __new__(cls, x: float, y: float): ...
    def __add__(self, other: Any) -> Any: ...
    def __mul__(self, other: Any) -> Any: ...
    def __rmul__(self, other: Any) -> Any: ...
    def __sub__(self, other: Any) -> Any: ...
    def __neg__(self) -> "Vec2D": ...
    def __abs__(self) -> "Vec2D": ...
    def rotate(self, angle: Any) -> "Vec2D": ...
    def __getnewargs__(self) -> Tuple[float, float]: ...

Point2D = Union[Vec2D, Tuple[float, float]]

class ScrolledCanvas(TK.Frame):
    bg = ""
    hscroll = TK.Scrollbar()
    vscroll = TK.Scrollbar()
    def __init__(
        self,
        master: TK.Widget,
        width: int = ...,
        height: int = ...,
        canvwidth: int = ...,
        canvheight: int = ...,
    ) -> None: ...
    canvwidth = 0
    canvheight = 0
    def reset(
        self,
        canvwidth: Optional[int] = ...,
        canvheight: Optional[int] = ...,
        bg: Optional[str] = ...,
    ) -> None: ...
    def adjustScrolls(self) -> None: ...
    def onResize(self, event: TK.Event) -> None: ...
    def bbox(self, *args: Any) -> Tuple[int, int, int, int]: ...
    def cget(self, *args: Any, **kwargs: Any): ...
    def config(self, *args: Any, **kwargs: Any) -> None: ...
    def bind(self, *args: Any, **kwargs: Any) -> None: ...
    def unbind(self, *args: Any, **kwargs: Any) -> None: ...
    def focus_force(self) -> None: ...

class _Root(TK.Tk):
    def __init__(self) -> None: ...
    def setupcanvas(self, width: Any, height: Any, cwidth: Any, cheight: Any) -> None: ...
    def set_geometry(self, width: Any, height: Any, startx: Any, starty: Any) -> None: ...
    def ondestroy(self, destroy: Any) -> None: ...
    def win_width(self): ...
    def win_height(self): ...

class TurtleScreenBase:
    cv = TK.Canvas()
    canvwidth = 0
    canvheight = 0
    xscale = 0.0
    def __init__(self, cv: TK.Canvas) -> None: ...
    def mainloop(self) -> None: ...
    def textinput(self, title: str, prompt: str) -> Optional[str]: ...
    def numinput(
        self,
        title: str,
        prompt: str,
        default: Optional[float] = ...,
        minval: Optional[float] = ...,
        maxval: Optional[float] = ...,
    ) -> Optional[float]: ...

class Terminator(Exception): ...
class TurtleGraphicsError(Exception): ...

class Shape:
    def __init__(self, type_: str, data: Optional[Any] = None) -> None: ...
    def addcomponent(self, poly: Any, fill: Any, outline: Optional[Any] = None) -> None: ...

class Tbuffer:
    bufsize = 0
    buffer = ...  # type: Any
    ptr = 0
    cumulate = True
    def __init__(self, bufsize: int = ...) -> None: ...
    def reset(self, bufsize: Optional[Any] = None) -> None: ...
    def push(self, item: Any) -> None: ...
    def pop(self): ...
    def nr_of_items(self): ...

class TurtleScreen(TurtleScreenBase):
    def __init__(self, cv: TK.Canvas, mode: Any, colormode: Any, delay: Any) -> None: ...
    def clear(self) -> None: ...
    xscale = 0.0
    def mode(self, mode: Optional[str] = ...) -> Optional[str]: ...
    yscale = ...  # type: Any
    def setworldcoordinates(self, llx: Any, lly: Any, urx: Any, ury: Any) -> None: ...
    def register_shape(self, name: Any, shape: Optional[Any] = None) -> None: ...
    def colormode(self, cmode: Optional[float] = ...) -> Optional[float]: ...
    def reset(self) -> None: ...
    def turtles(self) -> List["Turtle"]: ...
    def bgcolor(
        self, *args: Union[Tuple[str], Tuple[float, float, float]]
    ) -> Union[Tuple[str], Tuple[float, float, float]]: ...
    def tracer(self, n: Optional[Any] = None, delay: Optional[Any] = None): ...
    def delay(self, delay: Optional[Any] = None): ...
    def update(self) -> None: ...
    def window_width(self): ...
    def window_height(self): ...
    def getcanvas(self): ...
    def getshapes(self): ...
    def onclick(self, fun: Any, btn: int = ..., add: Optional[Any] = None) -> None: ...
    def onkey(self, fun: Any, key: Any) -> None: ...
    def onkeypress(self, fun: Any, key: Optional[Any] = None) -> None: ...
    def listen(self, xdummy: Optional[Any] = None, ydummy: Optional[Any] = None) -> None: ...
    def ontimer(self, fun: Any, t: int = ...) -> None: ...
    def bgpic(self, picname: Optional[Any] = None): ...
    def screensize(
        self,
        canvwidth: Optional[Any] = None,
        canvheight: Optional[Any] = None,
        bg: Optional[Any] = None,
    ): ...
    onscreenclick = onclick
    resetscreen = reset
    clearscreen = clear
    addshape = register_shape
    onkeyrelease = onkey

class TNavigator:
    START_ORIENTATION = {}  # type: Dict[str, Vec2D]
    DEFAULT_MODE = ""
    DEFAULT_ANGLEOFFSET = 0
    DEFAULT_ANGLEORIENT = 0
    undobuffer = ...  # type: Any
    def __init__(self, mode: str = ...) -> None: ...
    def reset(self) -> None: ...
    def degrees(self, fullcircle: float = ...) -> None: ...
    def radians(self) -> None: ...
    def forward(self, distance: float) -> None: ...
    def back(self, distance: float) -> None: ...
    def right(self, angle: float) -> None: ...
    def left(self, angle: float) -> None: ...
    def pos(self) -> Vec2D: ...
    def xcor(self) -> float: ...
    def ycor(self) -> float: ...
    @overload
    def goto(self, x: float, y: float) -> None: ...
    @overload
    def goto(self, x: Point2D) -> None: ...
    def home(self) -> None: ...
    def setx(self, x: float) -> None: ...
    def sety(self, y: float) -> None: ...
    @overload
    def distance(self, x: float, y: float) -> float: ...
    @overload
    def distance(self, x: Union[Turtle, Point2D]) -> float: ...
    @overload
    def towards(self, x: float, y: float) -> float: ...
    @overload
    def towards(self, x: Union[Turtle, Point2D]) -> float: ...
    def heading(self) -> float: ...
    def setheading(self, to_angle: float) -> None: ...
    def circle(
        self, radius: float, extent: Optional[float] = None, steps: Optional[float] = None
    ) -> None: ...
    def speed(self, s: int = ...) -> None: ...
    fd = forward
    bk = back
    backward = back
    rt = right
    lt = left
    position = pos
    setpos = goto
    setposition = goto
    seth = setheading

class TPen:
    undobuffer = ...  # type: Any
    def __init__(self, resizemode: Any) -> None: ...
    def resizemode(self, rmode: Optional[Any] = None): ...
    def pensize(self, width: Optional[Any] = None): ...
    def penup(self): ...
    def pendown(self): ...
    def isdown(self): ...
    def speed(self, speed: Optional[Any] = None): ...
    def color(self, *args: Any): ...
    def pencolor(self, *args: Any): ...
    def fillcolor(self, *args: Any): ...
    def showturtle(self) -> None: ...
    def hideturtle(self) -> None: ...
    def isvisible(self): ...
    def pen(self, pen: Optional[Any] = None, **pendict: Any): ...
    width = pensize
    up = penup
    pu = penup
    pd = pendown
    down = pendown
    st = showturtle
    ht = hideturtle

class _TurtleImage:
    screen = ...  # type: Any
    def __init__(self, screen: Any, shapeIndex: Any) -> None: ...

class RawTurtle(TPen, TNavigator):
    screens = ...  # type: Any
    screen = ...  # type: Any
    drawingLineItem = ...  # type: Any
    turtle = ...  # type: Any
    currentLineItem = ...  # type: Any
    currentLine = ...  # type: Any
    items = ...  # type: Any
    stampItems = ...  # type: Any
    undobuffer = ...  # type: Any
    def __init__(
        self,
        canvas: Optional[Any] = None,
        shape: Any = None,
        undobuffersize: Any = None,
        visible: Any = None,
    ) -> None: ...
    def reset(self) -> None: ...
    def setundobuffer(self, size: Any) -> None: ...
    def undobufferentries(self): ...
    def clear(self) -> None: ...
    def clone(self): ...
    def shape(self, name: Optional[Any] = None): ...
    def shapesize(
        self,
        stretch_wid: Optional[Any] = None,
        stretch_len: Optional[Any] = None,
        outline: Optional[Any] = None,
    ): ...
    def shearfactor(self, shear: Optional[Any] = None): ...
    def settiltangle(self, angle: Any) -> None: ...
    def tiltangle(self, angle: Optional[Any] = None): ...
    def tilt(self, angle: Any) -> None: ...
    def shapetransform(
        self,
        t11: Optional[Any] = None,
        t12: Optional[Any] = None,
        t21: Optional[Any] = None,
        t22: Optional[Any] = None,
    ): ...
    def get_shapepoly(self): ...
    def stamp(self): ...
    def clearstamp(self, stampid: Any) -> None: ...
    def clearstamps(self, n: Optional[Any] = None) -> None: ...
    def filling(self): ...
    def begin_fill(self) -> None: ...
    def end_fill(self) -> None: ...
    def dot(self, size: Optional[Any] = None, *color: Any) -> None: ...
    def write(self, arg: Any, move: bool = ..., align: str = ..., font: Any = None) -> None: ...
    def begin_poly(self) -> None: ...
    def end_poly(self) -> None: ...
    def get_poly(self): ...
    def getscreen(self): ...
    def getturtle(self): ...
    getpen = ...  # type: Any
    def onclick(self, fun: Any, btn: int = ..., add: Optional[Any] = None) -> None: ...
    def onrelease(self, fun: Any, btn: int = ..., add: Optional[Any] = None) -> None: ...
    def ondrag(self, fun: Any, btn: int = ..., add: Optional[Any] = None) -> None: ...
    def undo(self): ...
    turtlesize = shapesize

RawPen = RawTurtle

class _Screen(TurtleScreen):
    def __init__(self) -> None: ...
    def setup(self, width: Any, height: Any, startx: Any, starty: Any): ...
    def title(self, titlestring: Any) -> None: ...
    def bye(self) -> None: ...
    def exitonclick(self): ...

def Screen() -> _Screen: ...

class Turtle(RawTurtle):
    def __init__(self, shape: Any, undobuffersize: Any, visible: Any) -> None: ...

Pen = Turtle

def write_docstringdict(filename: str = ...) -> None: ...

_screen = Screen()
_turtle = Turtle()

addshape = _screen.addshape
bgcolor = _screen.bgcolor
bgpic = _screen.bgpic
bye = _screen.bye
clearscreen = _screen.clearscreen
colormode = _screen.colormode
delay = _screen.delay
exitonclick = _screen.exitonclick
getcanvas = _screen.getcanvas
getshapes = _screen.getshapes
listen = _screen.listen
mainloop = _screen.mainloop
mode = _screen.mode
numinput = _screen.numinput
onkey = _screen.onkey
onkeypress = _screen.onkeypress
onkeyrelease = _screen.onkeyrelease
onscreenclick = _screen.onscreenclick
ontimer = _screen.ontimer
register_shape = _screen.register_shape
resetscreen = _screen.resetscreen
screensize = _screen.screensize
setup = _screen.setup
setworldcoordinates = _screen.setworldcoordinates
textinput = _screen.textinput
title = _screen.title
tracer = _screen.tracer
turtles = _screen.turtles
update = _screen.update
window_height = _screen.window_height
window_width = _screen.window_width

done = mainloop

back = _turtle.back
backward = _turtle.backward
begin_fill = _turtle.begin_fill
begin_poly = _turtle.begin_poly
bk = _turtle.bk
circle = _turtle.circle
clear = _turtle.clear
clearstamp = _turtle.clearstamp
clearstamps = _turtle.clearstamps
clone = _turtle.clone
color = _turtle.color
degrees = _turtle.degrees
distance = _turtle.distance
dot = _turtle.dot
down = _turtle.down
end_fill = _turtle.end_fill
end_poly = _turtle.end_poly
fd = _turtle.fd
fillcolor = _turtle.fillcolor
filling = _turtle.filling
forward = _turtle.forward
get_poly = _turtle.get_poly
get_shapepoly = _turtle.get_shapepoly
getpen = _turtle.getpen
getscreen = _turtle.getscreen
getturtle = _turtle.getturtle
goto = _turtle.goto
heading = _turtle.heading
hideturtle = _turtle.hideturtle
home = _turtle.home
ht = _turtle.ht
isdown = _turtle.isdown
isvisible = _turtle.isvisible
left = _turtle.left
lt = _turtle.lt
onclick = _turtle.onclick
ondrag = _turtle.ondrag
onrelease = _turtle.onrelease
pd = _turtle.pd
pen = _turtle.pen
pencolor = _turtle.pencolor
pendown = _turtle.pendown
pensize = _turtle.pensize
penup = _turtle.penup
pos = _turtle.pos
position = _turtle.position
pu = _turtle.pu
radians = _turtle.radians
reset = _turtle.reset
resizemode = _turtle.resizemode
right = _turtle.right
rt = _turtle.rt
seth = _turtle.seth
setheading = _turtle.setheading
setpos = _turtle.setpos
setposition = _turtle.setposition
settiltangle = _turtle.settiltangle
setundobuffer = _turtle.setundobuffer
setx = _turtle.setx
sety = _turtle.sety
shape = _turtle.shape
shapesize = _turtle.shapesize
shapetransform = _turtle.shapetransform
shearfactor = _turtle.shearfactor
showturtle = _turtle.showturtle
speed = _turtle.speed
st = _turtle.st
stamp = _turtle.stamp
tilt = _turtle.tilt
tiltangle = _turtle.tiltangle
towards = _turtle.towards
turtlesize = _turtle.turtlesize
undo = _turtle.undo
undobufferentries = _turtle.undobufferentries
up = _turtle.up
width = _turtle.width
write = _turtle.write
xcor = _turtle.xcor
ycor = _turtle.ycor



import os

from thonny import get_workbench

_OPTION_NAME = "run.dock_user_windows"


def toggle_variable():
    var = get_workbench().get_variable(_OPTION_NAME)
    var.set(not var.get())
    update_environment()


def update_environment():
    os.environ["DOCK_USER_WINDOWS"] = str(get_workbench().get_option(_OPTION_NAME))


def on_window_appear(event):
    get_workbench().become_active_window(force=False)


def load_plugin():
    get_workbench().set_default(_OPTION_NAME, False)
    get_workbench().add_command(
        "toggle_dock_user_windows",
        "run",
        _("Dock user windows"),
        toggle_variable,
        flag_name=_OPTION_NAME,
        group=40,
    )
    update_environment()

    get_workbench().bind("UserWindowAppeared", on_window_appear, True)



def acos():
    pass


def acosh():
    pass


def asin():
    pass


def asinh():
    pass


def atan():
    pass


def atan2():
    pass


def atanh():
    pass


def ceil():
    pass


def copysign():
    pass


def cos():
    pass


def cosh():
    pass


def degrees():
    pass


e = 2.718282


def erf():
    pass


def erfc():
    pass


def exp():
    pass


def expm1():
    pass


def fabs():
    pass


def floor():
    pass


def fmod():
    pass


def frexp():
    pass


def gamma():
    pass


def isfinite():
    pass


def isinf():
    pass


def isnan():
    pass


def ldexp():
    pass


def lgamma():
    pass


def log():
    pass


def log10():
    pass


def log2():
    pass


def modf():
    pass


pi = 3.141593


def pow():
    pass


def radians():
    pass


def sin():
    pass


def sinh():
    pass


def sqrt():
    pass


def tan():
    pass


def tanh():
    pass


def trunc():
    pass



CONFIG = "./webrepl_cfg.py"
RC = "./boot.py"


def change_daemon():
    pass


def exists():
    pass


def get_daemon_status():
    pass


def getpass():
    pass


def input_choice():
    pass


def input_pass():
    pass


machine = None


def main():
    pass


os = None
sys = None



class AudioOut:
    ""

    def deinit():
        pass

    def pause():
        pass

    paused = None

    def play():
        pass

    playing = None

    def resume():
        pass

    def stop():
        pass


class Mixer:
    ""

    def deinit():
        pass

    def play():
        pass

    playing = None
    sample_rate = None

    def stop_voice():
        pass


class RawSample:
    ""

    def deinit():
        pass

    sample_rate = None


class WaveFile:
    ""
    bits_per_sample = None
    channel_count = None

    def deinit():
        pass

    sample_rate = None



import os.path
import sys

from setuptools import find_packages, setup

def recursive_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

if sys.version_info < (3, 5):
    raise RuntimeError("Thonny requires Python 3.5 or later")

setupdir = os.path.dirname(__file__)

with open(os.path.join(setupdir, "thonny", "VERSION"), encoding="ASCII") as f:
    version = f.read().strip()

requirements = []
for line in open(os.path.join(setupdir, "requirements.txt"), encoding="ASCII"):
    if line.strip() and not line.startswith("#"):
        requirements.append(line)

setup(
    name="thonny",
    version=version,
    description="Python IDE for beginners",
    long_description="Thonny is a simple Python IDE with features useful for learning programming. See https://thonny.org for more info.",
    url="https://thonny.org",
    author="Aivar Annamaa and others",
    author_email="thonny@googlegroups.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "License :: Freeware",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Education",
        "Topic :: Software Development",
        "Topic :: Software Development :: Debuggers",
        "Topic :: Text Editors",
    ],
    keywords="IDE education debugger",
    project_urls={
        "Source code": "https://github.com/thonny/thonny",
        "Bug tracker": "https://github.com/thonny/thonny/issues",
    },
    platforms=["Windows", "macOS", "Linux"],
    install_requires=requirements,
    python_requires=">=3.5",
    packages=find_packages(),
    package_data={
        "": ["VERSION", "res/*"] + recursive_files("thonny/locale"),
        "thonny.plugins.help": ["*.rst"],
        "thonny.plugins.pi": ["res/*.*"],
        "thonny.plugins.printing": ["*.html"],
        "thonny.plugins.micropython": ["api_stubs/*.*"],
        "thonny.plugins.circuitpython": ["api_stubs/*.*"],
        "thonny.plugins.microbit": ["api_stubs/*.*"],
        "thonny.plugins.esp": ["esp32_api_stubs/*.*", "esp_8266_api_stubs/*.*"],
        "thonny.plugins.mypy": ["typeshed_extras/*.pyi"],
    },
    entry_points={"gui_scripts": ["thonny = thonny:launch"]},
)



import serial
from thonny.plugins.micropython.connection import MicroPythonConnection, ConnectionFailedException
import threading
import time
from serial.serialutil import SerialException
import logging
import platform
import sys
from textwrap import dedent


class SerialConnection(MicroPythonConnection):
    def __init__(self, port, baudrate, skip_reader=False):
        super().__init__()

        try:
            self._serial = serial.Serial(port, baudrate=baudrate, timeout=None)
        except SerialException as error:
            err_str = str(error)
            if "FileNotFoundError" in err_str:
                err_str = "port not found"
            message = "Unable to connect to " + port + ": " + err_str

            # TODO: check if these error codes also apply to Linux and Mac
            if error.errno == 13 and platform.system() == "Linux":
                # TODO: check if user already has this group
                message += "\n\n" + dedent(
                    """\
                Try adding yourself to the 'dialout' group:
                > sudo usermod -a -G dialout <username>
                (NB! This needs to be followed by reboot or logging out and logging in again!)"""
                )

            elif "PermissionError" in message:
                message += "\n\n" + dedent(
                    """\
                If you have serial connection to the device from another program,
                then disconnect it there."""
                )

            elif error.errno == 16:
                message += "\n\n" + "Try restarting the device."

            raise ConnectionFailedException(message)

        if skip_reader:
            self._reading_thread = None
        else:
            self._reading_thread = threading.Thread(target=self._listen_serial, daemon=True)
            self._reading_thread.start()

    def write(self, data, block_size=256, delay=0.01):
        # delay and block size taken from rshell
        # https://github.com/dhylands/rshell/blob/master/rshell/pyboard.py#L242
        for i in range(0, len(data), block_size):
            block = data[i : i + block_size]
            # self._log_data(b"[" + block + b"]")
            size = self._serial.write(block)
            assert size == len(block)
            time.sleep(delay)
        return len(data)

    def _listen_serial(self):
        "NB! works in background thread"
        try:
            data = b""
            while True:
                data += self._serial.read(1)  # To avoid busy loop
                if len(data) == 0:
                    self._error = "EOF"
                    # print("LISTEN EOFFFFFFFFFF")
                    break
                data += self._serial.read_all()
                self.num_bytes_received += len(data)

                # don't publish incomplete utf-8 data
                try:
                    data.decode("utf-8")  # testing if data decodes
                    to_be_published = data
                    data = b""
                except UnicodeDecodeError as e:
                    if e.start == 0:
                        # Invalid start byte, ie. we have missed first byte(s) of the codepoint.
                        # No use of waiting, output everything
                        to_be_published = data
                        data = b""
                    else:
                        to_be_published = data[: e.start]
                        data = data[e.start :]

                if to_be_published:
                    self._read_queue.put(to_be_published)

        except Exception as e:
            self._error = str(e)

    def incoming_is_empty(self):
        return self._serial.in_waiting == 0 and super().incoming_is_empty()

    def outgoing_is_empty(self):
        return self._serial.out_waiting == 0

    def reset_output_buffer(self):
        self._serial.reset_output_buffer()

    def close(self):
        if self._serial is not None:
            try:
                self._serial.cancel_read()
                if self._reading_thread:
                    self._reading_thread.join()
            finally:
                try:
                    self._serial.close()
                    self._serial = None
                except Exception:
                    logging.exception("Couldn't close serial")


def debug(*args, file=sys.stderr):
    print(*args, file=file)



import platform
import shutil
import sys
import os.path


def _clear_screen():
    if platform.system() == "Windows":
        os.system("cls")
    else:
        os.system("clear")


def equivalent_realpath(p):
    pdir = os.path.dirname(p)
    if os.path.isfile(os.path.join(pdir, "activate")):
        # it's a virtual environment
        # can use realpath only if it doesn't move out of its dir
        real = os.path.realpath(p)
        if os.path.dirname(real) == pdir:
            return real
        try:
            link = os.readlink(p)
            if not os.path.isabs(link):
                link = os.path.join(pdir, link)
            link = os.path.normpath(link)
            if os.path.dirname(link) == pdir:
                return link
            return p
        except:
            return p
    else:
        return os.path.realpath(p)


def is_virtual_exe(p):
    pdir = os.path.dirname(p)
    return os.path.exists(os.path.join(pdir, "activate")) or os.path.exists(
        os.path.join(pdir, "activate.bat")
    )


def is_bundled_exe(p):
    pdir = os.path.dirname(p)
    return os.path.exists(os.path.join(pdir, "thonny_python.ini"))


def list_commands(prefix, highlighted_reals, highlighted_dirs):
    for suffix in ["", "3", "3.5", "3.6", "3.7", "3.8"]:
        cmd = prefix + suffix
        target = shutil.which(cmd)
        if target is not None:
            target = os.path.abspath(target)
            target = normpath_with_actual_case(target)
            real = equivalent_realpath(target)

            if target == real:
                relation = "=="
            else:
                relation = "->"

            line = " - " + cmd.ljust(9) + " " + relation + " " + real
            if (
                real in highlighted_reals
                or os.path.dirname(real) in highlighted_dirs
                or os.path.dirname(target) in highlight_dirs
            ):
                print("\033[1m" + line + "\033[0m")
            else:
                print("\033[2m" + line + "\033[0m")


def normpath_with_actual_case(name):
    """In Windows return the path with the case it is stored in the filesystem"""
    # copied from thonny.common to make this script independent
    assert os.path.isabs(name) or os.path.ismount(name), "Not abs nor mount: " + name
    assert os.path.exists(name), "Not exists: " + name

    if os.name == "nt":
        # https://stackoverflow.com/questions/2113822/python-getting-filename-case-as-stored-in-windows/2114975
        name = os.path.normpath(name)

        from ctypes import create_unicode_buffer, windll

        buf = create_unicode_buffer(512)
        # GetLongPathNameW alone doesn't fix filename part
        windll.kernel32.GetShortPathNameW(name, buf, 512)  # @UndefinedVariable
        windll.kernel32.GetLongPathNameW(buf.value, buf, 512)  # @UndefinedVariable
        result = buf.value

        if result.casefold() != name.casefold():
            # Sometimes GetShortPathNameW + GetLongPathNameW doesn't work
            # see eg. https://github.com/thonny/thonny/issues/925
            windll.kernel32.GetLongPathNameW(name, buf, 512)  # @UndefinedVariable
            result = buf.value

            if result.casefold() != name.casefold():
                result = name

        if result[1] == ":":
            # ensure drive letter is capital
            return result[0].upper() + result[1:]
        else:
            return result
    else:
        # easy on Linux
        # too difficult on mac
        # https://stackoverflow.com/questions/14515073/in-python-on-osx-with-hfs-how-can-i-get-the-correct-case-of-an-existing-filenam
        # Hopefully only correct case comes into Thonny (eg. via open dialog)
        return os.path.normpath(name)


if __name__ == "__main__":
    _clear_screen()
    print("*" * 80)
    print("Some Python commands in the PATH of this session:")

    sys_real = normpath_with_actual_case(equivalent_realpath(sys.executable))
    sys_executable = normpath_with_actual_case(sys.executable)

    if is_virtual_exe(sys_executable) or is_bundled_exe(sys_executable):
        highlight_dirs = [os.path.dirname(sys_executable)]
    else:
        highlight_dirs = []

    if platform.system() == "Windows":
        # Add Scripts for pip
        highlight_dirs.append(os.path.join(os.path.dirname(sys_real), "Scripts"))
        highlight_dirs.append(os.path.join(os.path.dirname(sys_executable), "Scripts"))

    list_commands("python", [sys_real], highlight_dirs)

    likely_pips = []
    if sys_real[-9:-1] == "python3.":
        likely_pips.append(sys_real[:-9] + "pip3." + sys_real[-1])
    if sys_executable.endswith("/python3"):
        # This is not as likely match as previous, but still quite likely
        likely_pips.append(sys_executable.replace("/python3", "/pip3"))

    list_commands("pip", likely_pips, highlight_dirs)

    print("")
    print("*" * 80)
